"""Image metrics.

The key difference between these implementations and those in torchmetrics
is that these metrics support operations on complex data types.
"""

from typing import Optional, Sequence, Union

import numpy as np
import torch
import torch.nn.functional as F
from torchmetrics.functional.image.ssim import _gaussian

from meddlr.ops import complex as cplx

__all__ = ["mse", "rmse", "psnr", "nrmse", "l2_norm", "ssim"]


# Mapping from str to complex function name.
_IM_TYPES_TO_FUNCS = {
    "mag": cplx.abs,
    "magnitude": cplx.abs,
    "abs": cplx.abs,
    "phase": cplx.angle,
    "angle": cplx.angle,
    "real": cplx.real,
    "imag": cplx.imag,
}


def _check_consistent_type(*args):
    is_complex = [cplx.is_complex(x) or cplx.is_complex_as_real(x) for x in args]
    all_complex = all(is_complex)
    all_not_complex = all(not x for x in is_complex)

    if not all_complex and not all_not_complex:
        raise ValueError("Type mismatch - all inputs must be complex or real")


def mse(pred: torch.Tensor, target: torch.Tensor, im_type: str = None) -> torch.Tensor:
    """Computes mean square error.

    Args:
        pred (torch.Tensor): The prediction. Either a complex or real tensor.
        target (torch.Tensor): The target. Either a complex or real tensor.
        im_type (str, optional): The image type to compute metric on.
            This only applies for complex inputs, otherwise ignored.
            Either ``'magnitude'`` (default) to compute metric on magnitude images
            or ``'phase'`` to compute metric on phase/angle images. If ``None``,
            computed on complex images.

    Returns:
        torch.Tensor: The mean square error.
    """
    if im_type is not None:
        pred = _IM_TYPES_TO_FUNCS[im_type](pred)
        target = _IM_TYPES_TO_FUNCS[im_type](target)

    if cplx.is_complex(pred) or cplx.is_complex_as_real(pred):
        squared_err = cplx.abs(pred - target) ** 2
    else:
        squared_err = torch.abs(pred - target) ** 2
    shape = (pred.shape[0], pred.shape[1], -1)
    return torch.mean(squared_err.view(shape), dim=-1)


def rmse(pred: torch.Tensor, target: torch.Tensor, im_type: str = None) -> torch.Tensor:
    """Computes root mean square error.

    Args:
        pred (torch.Tensor): The prediction. Either a complex or real tensor.
        target (torch.Tensor): The target. Either a complex or real tensor.
        im_type (str, optional): The image type to compute metric on.
            This only applies for complex inputs, otherwise ignored.
            Either ``'magnitude'`` (default) to compute metric on magnitude images
            or ``'phase'`` to compute metric on phase/angle images. If ``None``,
            computed on complex images.

    Returns:
        torch.Tensor: The root mean square error.
    """
    return torch.sqrt(mse(pred, target, im_type=im_type))


def psnr(pred: torch.Tensor, target: torch.Tensor, im_type: str = None) -> torch.Tensor:
    """Computes peak signal-to-noise ratio.

    Args:
        pred (torch.Tensor): The prediction. Either a complex or real tensor.
        target (torch.Tensor): The target. Either a complex or real tensor.
        im_type (str, optional): The image type to compute metric on.
            This only applies for complex inputs, otherwise ignored.
            Either ``'magnitude'`` (default) to compute metric on magnitude images
            or ``'phase'`` to compute metric on phase/angle images. If ``None``,
            computed on complex images.

    Returns:
        torch.Tensor: The peak signal-to-noise ratio.
    """
    is_complex = cplx.is_complex(pred) or cplx.is_complex_as_real(pred)
    abs_func = cplx.abs if is_complex else torch.abs

    l2_val = rmse(pred, target, im_type=im_type)
    shape = (target.shape[0], target.shape[1], -1)
    max_val = torch.amax(abs_func(target).view(shape), dim=-1)
    return 20 * torch.log10(max_val / l2_val)


def nrmse(pred: torch.Tensor, target: torch.Tensor, im_type: str = None) -> torch.Tensor:
    """Computes normalized root mean squared error.

    Normalization is done with respect to :math:`\sqrt{\\frac{\sum^N target[i]^2}{N}}`.

    Args:
        pred (torch.Tensor): The prediction. Either a complex or real tensor.
        target (torch.Tensor): The target. Either a complex or real tensor.
        im_type (str, optional): The image type to compute metric on.
            This only applies for complex inputs, otherwise ignored.
            Either ``'magnitude'`` (default) to compute metric on magnitude images
            or ``'phase'`` to compute metric on phase/angle images. If ``None``,
            computed on complex images.

    Returns:
        torch.Tensor: The normalized root mean squared error.
    """
    is_complex = cplx.is_complex(pred) or cplx.is_complex_as_real(pred)
    abs_func = cplx.abs if is_complex else torch.abs

    rmse_val = rmse(pred, target, im_type=im_type)
    shape = (pred.shape[0], pred.shape[1], -1)
    norm = torch.sqrt(torch.mean((abs_func(target) ** 2).view(shape), dim=-1))
    return rmse_val / norm


def l2_norm(pred: torch.Tensor, target: torch.Tensor) -> torch.Tensor:
    """Computes l2-norm of the error.

    Normalization is done with respect to :math:`\sqrt{\\frac{\sum^N target[i]^2}{N}}`.

    Args:
        pred (torch.Tensor): The prediction. Either a complex or real tensor.
        target (torch.Tensor): The target. Either a complex or real tensor.
        im_type (str, optional): The image type to compute metric on.
            This only applies for complex inputs, otherwise ignored.
            Either ``'magnitude'`` (default) to compute metric on magnitude images
            or ``'phase'`` to compute metric on phase/angle images. If ``None``,
            computed on complex images.

    Returns:
        torch.Tensor: The normalized root mean squared error.
    """
    err = pred - target
    is_complex = cplx.is_complex(err) or cplx.is_complex_as_real(err)
    abs_func = cplx.abs if is_complex else torch.abs
    shape = (pred.shape[0], pred.shape[1], -1)
    return torch.sum(abs_func(err).view(shape), dim=-1)


def ssim(
    pred: torch.Tensor,
    target: torch.Tensor,
    method: str = None,
    kernel_size=11,
    sigma=1.5,
    data_range=None,
    k1=0.01,
    k2=0.03,
    pad_mode: str = "reflect",
    im_type: str = "magnitude",
) -> torch.Tensor:
    """Computes structural similarity index (SSIM).

    Args:
        pred (torch.Tensor): The prediction. Either a complex or real tensor.
        target (torch.Tensor): The target. Either a complex or real tensor.
        im_type (str, optional): The image type to compute metric on.
            This only applies for complex inputs, otherwise ignored.
            Either ``'magnitude'`` (default) to compute metric on magnitude images
            or ``'phase'`` to compute metric on phase/angle images. If ``None``,
            computed on complex images.

    Returns:
        torch.Tensor: The SSIM for each (batch, channel) pair.
    """
    if method is not None:
        if method.lower() == "wang":
            kernel_size = 11
            sigma = 1.5
            data_range = "ref-maxval"
            k1 = 0.01
            k2 = 0.03
        else:
            raise ValueError(f"Unknown method {method}")

    if im_type is not None:
        if cplx.is_complex(pred) or cplx.is_complex_as_real(pred):
            pred = _IM_TYPES_TO_FUNCS[im_type](pred)
        if cplx.is_complex(target) or cplx.is_complex_as_real(target):
            target = _IM_TYPES_TO_FUNCS[im_type](target)

    ssim_idx = _ssim_compute(
        pred,
        target,
        kernel_size=kernel_size,
        sigma=sigma,
        data_range=data_range,
        k1=k1,
        k2=k2,
        pad_mode=pad_mode,
    )

    reduce_dims = tuple(range(2, pred.ndim))
    return ssim_idx.mean(reduce_dims)


def _ssim_compute(
    pred: torch.Tensor,
    target: torch.Tensor,
    kernel_size: Union[int, Sequence[int]] = 11,
    sigma: Sequence[float] = 1.5,
    data_range: Optional[Union[float, torch.Tensor]] = None,
    k1: float = 0.01,
    k2: float = 0.03,
    pad_mode="reflect",
) -> torch.Tensor:
    """Compute structural similarity.

    Args:
        pred (torch.Tensor): The prediction. Shape: ``BxCxHxW`` or ``BxCxDxHxW``.
        target (torch.Tensor): The target. Shape: ``BxCxHxW`` or ``BxCxDxHxW``.
        kernel_size (int | Sequence[int]): The kernel size. If this is a scalar,
            the same size will be used for all spatial dimensions. If this is
            sequence, it should follow the same spatial ordering as ``pred`` and
            ``target``.
    """
    if isinstance(kernel_size, int):
        kernel_size = (kernel_size,) * (pred.ndim - 2)
    if isinstance(sigma, (int, float)):
        sigma = (sigma,) * (pred.ndim - 2)
    if len(kernel_size) != pred.ndim - 2:
        raise ValueError(
            f"Expected `kernel_size` to be an integer or sequence of length equal to the "
            f"number of spatial dimensions. Got {kernel_size}."
        )
    if len(sigma) != pred.ndim - 2:
        raise ValueError(
            f"Expected `sigma` to be an integer or sequence of length equal to the "
            f"number of spatial dimensions. Got {sigma}."
        )

    if any(x % 2 == 0 or x <= 0 for x in kernel_size):
        raise ValueError(f"Expected `kernel_size` to have odd positive number. Got {kernel_size}.")

    if any(y <= 0 for y in sigma):
        raise ValueError(f"Expected `sigma` to have positive number. Got {sigma}.")

    reduce_dims = tuple(range(2, pred.ndim))
    if data_range in ("range", "ref-range"):
        data_range = torch.amax(target, dim=reduce_dims) - torch.amin(target, dim=reduce_dims)
    elif data_range in ("ref-maxval", "maxval"):
        data_range = torch.amax(target, dim=reduce_dims)
    elif data_range == "x-range":
        data_range = torch.amax(pred, dim=reduce_dims) - torch.amin(pred, dim=reduce_dims)
    elif data_range == "x-maxval":
        data_range = torch.amax(pred, dim=reduce_dims)
    elif data_range is None:
        data_range = torch.amax(
            torch.cat(
                [
                    torch.amax(pred, dim=reduce_dims) - torch.amin(pred, dim=reduce_dims),
                    torch.amax(target, dim=reduce_dims) - torch.amin(target, dim=reduce_dims),
                ],
                dim=0,
            ),
            dim=0,
        )
    if not isinstance(data_range, torch.Tensor):
        data_range = torch.as_tensor(data_range)

    ndim = len(kernel_size)
    c1 = (k1 * data_range) ** 2
    c2 = (k2 * data_range) ** 2
    c1 = c1.view(c1.shape + (1,) * (2 + ndim - c1.ndim))
    c2 = c2.view(c2.shape + (1,) * (2 + ndim - c2.ndim))
    device = pred.device

    channel = pred.size(1)
    dtype = pred.dtype
    kernel = _gaussian_kernel(channel, kernel_size, sigma, dtype, device)
    padding = tuple(
        pad for pad_set in [((k - 1) // 2,) * 2 for k in kernel_size[::-1]] for pad in pad_set
    )  # (pad_w, pad_w, pad_h, pad_h, ...)

    pred = _pad(pred, padding, mode=pad_mode)
    target = _pad(target, padding, mode=pad_mode)

    input_list = torch.cat(
        (pred, target, pred * pred, target * target, pred * target)
    )  # (5 * B, C, H, W)
    if ndim == 2:
        outputs = F.conv2d(input_list, kernel, groups=channel)
    else:
        outputs = F.conv3d(input_list, kernel, groups=channel)
    output_list = [outputs[x * pred.size(0) : (x + 1) * pred.size(0)] for x in range(len(outputs))]

    mu_pred_sq = output_list[0].pow(2)
    mu_target_sq = output_list[1].pow(2)
    mu_pred_target = output_list[0] * output_list[1]

    sigma_pred_sq = output_list[2] - mu_pred_sq
    sigma_target_sq = output_list[3] - mu_target_sq
    sigma_pred_target = output_list[4] - mu_pred_target

    upper = 2 * sigma_pred_target + c2
    lower = sigma_pred_sq + sigma_target_sq + c2

    ssim_idx = ((2 * mu_pred_target + c1) * upper) / ((mu_pred_sq + mu_target_sq + c1) * lower)
    if ndim == 2:
        ssim_idx = ssim_idx[..., padding[2] : -padding[3], padding[0] : -padding[1]]
    else:
        ssim_idx = ssim_idx[
            ..., padding[4] : -padding[5], padding[2] : -padding[3], padding[0] : -padding[1]
        ]

    return ssim_idx


def _gaussian_kernel(
    channel: int,
    kernel_size: Sequence[int],
    sigma: Sequence[float],
    dtype: torch.dtype,
    device: torch.device,
) -> torch.Tensor:
    ndim = len(kernel_size)

    gaussian_kernel_x = _gaussian(kernel_size[0], sigma[0], dtype, device)
    gaussian_kernel_y = _gaussian(kernel_size[1], sigma[1], dtype, device)
    kernel = torch.matmul(
        gaussian_kernel_x.t(), gaussian_kernel_y
    )  # (kernel_size, 1) * (1, kernel_size)
    if ndim == 3:
        gaussian_kernel_z = _gaussian(kernel_size[2], sigma[2], dtype, device)
        kernel = kernel.unsqueeze(0) * gaussian_kernel_z.t().unsqueeze(-1)

    return kernel.expand(channel, 1, *kernel_size)


def _pad(x, padding, mode):
    if x.ndim < 5 or mode != "reflect":
        return F.pad(x, padding, mode)

    assert x.ndim == 5

    # 3D reflection padding
    # TODO: This will likely be supported in future PyTorch versions.
    # Update when the support is in a stable release.
    # https://github.com/pytorch/pytorch/pull/59791
    x = _pad_3d_tensor_with_2d_padding(x, padding[:-2], mode)

    dim = 2
    dpad1, dpad2 = padding[-2:]
    x1 = torch.flip(x[:, :, 1 : dpad1 + 1, ...], dims=(dim,))
    x2 = torch.flip(x[:, :, -dpad2 - 1 : -1, ...], dims=(dim,))
    x = torch.cat([x1, x, x2], dim=dim)
    return x


def _pad_3d_tensor_with_2d_padding(x, padding, mode):
    shape = x.shape
    x = x.reshape(shape[0], np.prod(shape[1:-2]), shape[-2], shape[-1])  # B x C*D x H x W
    x = F.pad(x, padding, mode)
    x = x.reshape(shape[0], *shape[1:-2], x.shape[-2], x.shape[-1])  # B x C x D x Hp x Wp
    return x



from typing import Sequence, Union

import numpy as np
import torch

from meddlr.ops import complex as cplx
from meddlr.utils import env
from meddlr.utils.events import get_event_storage

if env.pt_version() >= [1, 6]:
    import torch.fft


class NoiseAndMotionModel:
    """A model that adds additive white noise after adding simple motion. N(M(x))"""

    def __init__(
        self,
        std_devs: Union[float, Sequence[float]],
        motion_range: Union[float, Sequence[float]],
        scheduler=None,
        seed=None,
        device=None,
    ):
        if not isinstance(std_devs, Sequence):
            std_devs = (std_devs,)
        if isinstance(motion_range, (float, int)):
            motion_range = (motion_range,)
        elif len(std_devs) > 2:
            raise ValueError("`std_devs` must have 2 or fewer values")
        elif len(motion_range) > 2:
            raise ValueError("`motion_range` must have 2 or fewer values")
        self.std_devs = std_devs
        self.motion_range = motion_range

        self.warmup_method = None
        self.warmup_iters = 0
        if scheduler is not None:
            self.warmup_method = scheduler.WARMUP_METHOD
            self.warmup_iters = scheduler.WARMUP_ITERS

        # For reproducibility.
        g = torch.Generator(device=device)
        if seed:
            g = g.manual_seed(seed)
        self.generator = g

    def choose_std_dev(self):
        """Chooses std range based on warmup."""
        if not isinstance(self.std_devs, Sequence):
            return self.std_devs
        elif len(self.std_devs) == 1:
            return self.std_devs[0]

        if self.warmup_method:
            curr_iter = get_event_storage().iter
            warmup_iters = self.warmup_iters
            if self.warmup_method == "linear":
                std_range = curr_iter / warmup_iters * (self.std_devs[1] - self.std_devs[0])
            else:
                raise ValueError(f"`warmup_method={self.warmup_method}` not supported")
        else:
            std_range = self.std_devs[1] - self.std_devs[0]

        g = self.generator
        std_dev = self.std_devs[0] + std_range * torch.rand(1, generator=g, device=g.device).item()
        return std_dev

    def choose_motion_range(self):
        """Chooses motion range based on warmup."""
        if not isinstance(self.motion_range, Sequence):
            return self.motion_range
        elif len(self.motion_range) == 1:
            return self.motion_range[0]

        if self.warmup_method:
            curr_iter = get_event_storage().iter
            warmup_iters = self.warmup_iters
            if self.warmup_method == "linear":
                motion_range = (
                    curr_iter / warmup_iters * (self.motion_range[1] - self.motion_range[0])
                )
            else:
                raise ValueError(f"`warmup_method={self.warmup_method}` not supported")
        else:
            motion_range = self.motion_range[1] - self.motion_range[0]

        g = self.generator
        motion_range = (
            self.motion_range[0] + motion_range * torch.rand(1, generator=g, device=g.device).item()
        )
        return motion_range

    def __call__(self, *args, **kwargs):
        return self.forward(*args, **kwargs)

    def forward(self, kspace, mask=None, seed=None, clone=True) -> torch.Tensor:
        """Performs noise augmentation followed by motion simulation on undersampled kspace mask."""
        if clone:
            kspace = kspace.clone()
        mask = cplx.get_mask(kspace)

        g = (
            self.generator
            if seed is None
            else torch.Generator(device=kspace.device).manual_seed(seed)
        )

        phase_matrix = torch.zeros(kspace.shape, dtype=torch.complex64)
        width = kspace.shape[2]

        noise_std = self.choose_std_dev()
        scale = self.choose_motion_range()

        odd_err = (2 * np.pi * scale) * torch.rand(1, generator=g).numpy() - np.pi * scale
        even_err = (2 * np.pi * scale) * torch.rand(1, generator=g).numpy() - np.pi * scale

        for line in range(width):
            if line % 2 == 0:
                rand_err = even_err
            else:
                rand_err = odd_err
            phase_error = torch.from_numpy(np.exp(-1j * rand_err))
            phase_matrix[:, :, line] = phase_error
        aug_kspace = kspace * phase_matrix

        if cplx.is_complex(aug_kspace):
            noise = noise_std * torch.randn(
                aug_kspace.shape + (2,), generator=g, device=aug_kspace.device
            )
            noise = torch.view_as_complex(noise)
        else:
            noise = noise_std * torch.randn(aug_kspace.shape, generator=g, device=aug_kspace.device)
        masked_noise = noise * mask
        noised_aug_kspace = aug_kspace + masked_noise

        return noised_aug_kspace

    @classmethod
    def from_cfg(cls, cfg, seed=None, **kwargs):
        cfg = cfg.MODEL.CONSISTENCY.AUG
        return cls(
            cfg.NOISE.STD_DEV, cfg.MOTION.RANGE, scheduler=cfg.NOISE.SCHEDULER, seed=seed, **kwargs
        )



from typing import Sequence, Tuple

import torch

from meddlr.transforms.base.noise import NoiseTransform
from meddlr.transforms.build import TRANSFORM_REGISTRY
from meddlr.transforms.transform import NoOpTransform
from meddlr.transforms.transform_gen import TransformGen

__all__ = ["RandomNoise"]


@TRANSFORM_REGISTRY.register()
class RandomNoise(TransformGen):
    """A model that adds additive white noise."""

    _base_transform = NoiseTransform
    _param_names = ["std_devs", "rhos"]

    def __init__(
        self,
        p: float = 0.0,
        std_devs: Tuple[float, float] = None,
        rhos: Tuple[float, float] = None,
        use_mask: bool = True,
    ):
        std_devs = (std_devs, std_devs) if not isinstance(std_devs, Sequence) else std_devs
        if rhos is not None:
            rhos = (rhos, rhos) if not isinstance(rhos, Sequence) else rhos

        params = locals()
        params = {k: params[k] for k in list(self._param_names) + ["p"]}
        self.use_mask = use_mask
        super().__init__(params=params, p=p)

    def get_transform(self, input: torch.Tensor):
        params = self._get_param_values(use_schedulers=True)
        std_devs = params["std_devs"]
        rho = params["rhos"]

        if self._rand() >= params["p"]:
            return NoOpTransform()

        std_dev = self._rand_range(*std_devs)
        if rho is not None:
            rho = self._rand_range(*rho)

        gen = self._generator
        if gen is None or gen.device != input.device:
            gen = torch.Generator(device=input.device).manual_seed(int(self._rand() * 1e10))
        return NoiseTransform(std_dev=std_dev, use_mask=self.use_mask, rho=rho, generator=gen)



import warnings
from typing import Sequence, Tuple

import torch

from meddlr.ops import complex as cplx
from meddlr.transforms.build import TRANSFORM_REGISTRY
from meddlr.transforms.transform import Transform


@TRANSFORM_REGISTRY.register()
class KspaceMaskTransform(Transform):
    """A deterministic transform that masks kspace.

    This transform masks either fully-sampled or already undersampled kspace.
    Samples from kspace that are non-zero are dropped with probability defined
    by the kind of masking. ``rho`` is the fraction of pixels to drop.

    Uniform masking drops all non-zero pixels with a uniform probability.
    Gaussian masking drops all non-zero pixels with a probability modulated
    by the gaussian distribution centered at the center of the spatial
    kspace dimensions.

    Attributes:
        rho (float): The fraction of non-zero pixels to drop.
        kind (str): The kind of mask to use.
        std_scale (float): The standard deviation of the gaussian mask.
        per_example (bool): Whether to apply the mask per example
            separately or over the full batch. Defaults to False.
        calib_size (int, Tuple[int]): The size of the calibration region to use.
        seed (int): The seed to use for the random number generator.
        generator (torch.Generator): The random number generator to use.

    Note:
        This function is currently limited to 2D undersampling.

    Note:
        For this transform to be deterministic, either ``seed`` or ``generator``
        have to be set.

    Examples:
        >>> kspace = torch.randn(2, 100, 100, 8)
        >>> masker = KspaceMaskTransform(rho=0.5, kind='uniform', seed=42)
        >>> masked_kspace = masker(kspace)
        >>> mask = masker.generate_mask(kspace)  # to only generate mask
    """

    _SUPPORTED_MASK_KINDS = ("uniform", "gaussian")

    def __init__(
        self,
        rho: float,
        kind="uniform",
        std_scale=4.0,
        per_example=False,
        calib_size=None,
        seed: int = None,
        generator: torch.Generator = None,
    ):
        """
        Args:
            rho (float): The fraction of non-zero pixels to drop.
            kind (str, optional): The kind of mask to use. One of ``'uniform'`` or ``'gaussian'``.
            std_scale (float, optional): The standard deviation of the gaussian mask.
            per_example (bool, optional): Whether to apply the mask per example
                separately or over the full batch. Defaults to False.
            calib_size (int or Tuple[int], optional): The size of the calibration region to use.
            seed (int, optional): The seed to use for the random number generator.
            generator (torch.Generator, optional): The random number generator to use.
        """
        if kind not in self._SUPPORTED_MASK_KINDS:
            raise ValueError(
                f"Unknown kspace mask kind={kind}. " f"Expected one of {self._SUPPORTED_MASK_KINDS}"
            )
        self.rho = rho
        self.kind = kind
        self.seed = seed
        self.calib_size = calib_size
        self.per_example = per_example
        self.std_scale = std_scale

        gen_state = None
        if generator is not None:
            gen_state = generator.get_state()
        self._generator_state = gen_state

    def generate_mask(
        self, kspace: torch.Tensor, mask: torch.Tensor = None, channels_last: bool = False
    ):
        """Generates mask with approximiately ``1-self.rho`` times number of pixels.

        Args:
            kspace (torch.Tensor): The batch of kspace to generate the mask for.
            channels_last (bool, optional): If ``True``, the kspace is
                of shape ``[B, H, W, C]``. Defaults to ``False``.

        Returns:
            torch.Tensor: The generated mask.
        """
        g = self._generator(kspace)
        if mask is None:
            mask = True

        if channels_last:
            kspace = cplx.channels_first(kspace)
            if isinstance(mask, torch.Tensor):
                order = (0, mask.ndim - 1) + tuple(range(1, mask.ndim - 1))
                mask = mask.permute(order)

        func_and_kwargs = {
            "uniform": (
                _uniform_mask,
                {"rho": self.rho, "mask": mask, "calib_size": self.calib_size, "generator": g},
            ),
            "gaussian": (
                _gaussian_mask,
                {
                    "rho": self.rho,
                    "mask": mask,
                    "std_scale": self.std_scale,
                    "calib_size": self.calib_size,
                    "generator": g,
                },
            ),
        }
        func, kwargs = func_and_kwargs[self.kind]

        if self.per_example:
            mask = torch.cat([func(kspace[i : i + 1], **kwargs) for i in range(len(kspace))], dim=0)
        else:
            mask = func(kspace, **kwargs)

        if channels_last:
            order = (0,) + tuple(range(2, mask.ndim)) + (1,)
            mask = mask.permute(order)
        return mask

    def apply_kspace(self, kspace: torch.Tensor, channels_last: bool = False):
        """Apply transform to kspace.

        Args:
            kspace (torch.Tensor): The batch of kspace to generate the mask for.
            channels_last (bool, optional): If ``True``, the kspace is
                of shape ``[B, H, W, C]``. Defaults to ``False``.

        Returns:
            torch.Tensor: The masked kspace.
        """
        mask = self.generate_mask(kspace, channels_last=channels_last)
        return mask * kspace

    def _generator(self, data):
        seed = self.seed

        g = torch.Generator(device=data.device)
        if seed is None:
            g.set_state(self._generator_state)
        else:
            g = g.manual_seed(seed)
        return g

    def _subsample(self, data: torch.Tensor):  # pragma: no cover
        klass_name = {type(self).__name__}
        warnings.warn(
            f"{klass_name}._subsample is deprecated. Use {klass_name}.apply_kspace instead",
            DeprecationWarning,
        )

        mask = self.generate_mask(data)
        return mask * data

    def _eq_attrs(self) -> Tuple[str]:
        return ("std_dev", "use_mask", "rho", "seed", "_generator_state")


def _uniform_mask(kspace, rho, mask=True, calib_size=None, generator=None):
    """
    Args:
        kspace (torch.Tensor): The kspace to mask. Shape: ``BxCxHxWx...``.
        rho (float): Fraction of samples to drop.
    """
    kspace = kspace[:, 0:1, ...]
    orig_mask = cplx.get_mask(kspace) if mask is True else mask
    shape = orig_mask.shape
    ndim = orig_mask.ndim

    mask = orig_mask.clone()

    calib_region = None
    if calib_size is not None:
        if not isinstance(calib_size, Sequence):
            calib_size = (calib_size,) * (ndim - 2)
        center = tuple(s // 2 for s in shape[2:])[-len(calib_size) :]
        calib_region = tuple(slice(s - cs // 2, s + cs // 2) for s, cs in zip(center, calib_size))
        calib_region = (Ellipsis,) + calib_region
        mask[calib_region] = 0

    mask = mask.view(-1) if mask.is_contiguous() else mask.reshape(-1)

    # TODO: this doesnt work if the matrix is > 2*24 in size.
    # TODO: make this a bit more optimized
    num_valid = torch.sum(mask)
    weights = mask / num_valid
    num_samples = int(rho * num_valid)
    samples = torch.multinomial(weights, num_samples, replacement=False, generator=generator)
    mask[samples] = 0

    mask = mask.view(shape)
    if calib_region:
        mask[calib_region] = orig_mask[calib_region]
    return mask


def _gaussian_mask(kspace, rho, std_scale, mask=True, calib_size=None, generator=None):
    """
    Args:
        rho (float): Fraction of samples to drop.

    Note:
        Currently this creates the same mask for all the examples.
        Make sure to use per_example=True in :cls:`KspaceMaskTransform`.

    Note:
        This method is currently very slow.
    """
    kspace = kspace[:, 0:1, ...]
    shape = kspace.shape

    if mask is True:
        orig_mask = cplx.get_mask(kspace)
    else:
        orig_mask = mask
    mask = orig_mask.clone()

    spatial_shape = kspace.shape[2:]
    center = tuple(s // 2 for s in spatial_shape)

    calib_region = None
    if calib_size is not None:
        if not isinstance(calib_size, Sequence):
            calib_size = (calib_size,) * (kspace.ndim - 2)
        calib_region = tuple(
            slice(s - cs // 2, s + cs // 2) for s, cs in zip(center[-len(calib_size) :], calib_size)
        )
        calib_region = (Ellipsis,) + calib_region
        mask[calib_region] = 0

    num_valid = torch.sum(mask)
    num_samples = int(rho * num_valid)

    temp_mask = torch.zeros_like(mask)
    count = 0
    while count < num_samples:
        idxs = [
            torch.round(torch.normal(float(c), (s - 1) / std_scale, (1,))).type(torch.long)
            for c, s in zip(center, spatial_shape)
        ]
        if any(i < 0 for i in idxs) or any(i >= s for i, s in zip(idxs, spatial_shape)):
            continue
        idxs.insert(0, Ellipsis)
        if torch.all(mask[idxs] == 0) or torch.all(temp_mask[idxs] == 1):
            continue
        temp_mask[idxs] = 1
        count += 1

    mask = mask - temp_mask

    mask = mask.view(shape)
    if calib_region:
        mask[calib_region] = orig_mask[calib_region]
    return mask



import importlib
import importlib.util
import logging
import multiprocessing as mp
import os
import random
import re
import subprocess
import sys
import warnings
from datetime import datetime
from importlib import util
from typing import List

import numpy as np
import torch
from iopath.common.file_io import PathManager, PathManagerFactory
from packaging import version

__all__ = []

_PT_VERSION = torch.__version__
_SETTINGS_DIR = os.path.abspath(os.path.join(os.path.dirname(__file__), "../.settings"))
_GITHUB_URL = "https://github.com/ad12/meddlr"
_SUPPORTED_PACKAGES = {}


def seed_all_rng(seed=None):
    """
    Set the random seed for the RNG in torch, numpy and python.

    Args:
        seed (int): if None, will use a strong random seed.

    Returns:
        seed (int): The seed used.
    """
    if seed is None:
        seed = (
            os.getpid()
            + int(datetime.now().strftime("%S%f"))
            + int.from_bytes(os.urandom(2), "big")
        )
        logger = logging.getLogger(__name__)
        logger.info("Using a generated random seed {}".format(seed))
    np.random.seed(seed)
    torch.set_rng_state(torch.manual_seed(seed).get_state())
    random.seed(seed)
    return seed


# from https://stackoverflow.com/questions/67631/how-to-import-a-module-given-the-full-path  # noqa
def _import_file(module_name, file_path, make_importable=False):
    spec = importlib.util.spec_from_file_location(module_name, file_path)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    if make_importable:
        sys.modules[module_name] = module
    return module


def _configure_libraries():
    """
    Configurations for some libraries.
    """
    # An environment option to disable `import cv2` globally,
    # in case it leads to negative performance impact
    disable_cv2 = int(os.environ.get("MEDDLR_DISABLE_CV2", False))
    if disable_cv2:
        sys.modules["cv2"] = None
    else:
        # Disable opencl in opencv since its interaction with cuda often
        # has negative effects
        # This envvar is supported after OpenCV 3.4.0
        os.environ["OPENCV_OPENCL_RUNTIME"] = "disabled"
        try:
            import cv2

            if int(cv2.__version__.split(".")[0]) >= 3:
                cv2.ocl.setUseOpenCL(False)
        except ImportError:
            pass


_ENV_SETUP_DONE = False


def setup_environment():
    """Perform environment setup work. The default setup is a no-op, but this
    function allows the user to specify a Python source file or a module in
    the $MEDSEGPY_ENV_MODULE environment variable, that performs
    custom setup work that may be necessary to their computing environment.
    """
    global _ENV_SETUP_DONE
    if _ENV_SETUP_DONE:
        return
    _ENV_SETUP_DONE = True

    _configure_libraries()

    custom_module_path = os.environ.get("MEDDLR_ENV_MODULE")

    if custom_module_path:
        setup_custom_environment(custom_module_path)
    else:
        # The default setup is a no-op
        pass


def setup_custom_environment(custom_module):
    """
    Load custom environment setup by importing a Python source file or a
    module, and run the setup function.
    """
    if custom_module.endswith(".py"):
        module = _import_file("meddlr.utils.env.custom_module", custom_module)
    else:
        module = importlib.import_module(custom_module)
    assert hasattr(module, "setup_environment") and callable(module.setup_environment), (
        "Custom environment module defined in {} does not have the "
        "required callable attribute 'setup_environment'."
    ).format(custom_module)
    module.setup_environment()


def get_available_gpus(num_gpus: int = None):
    """Get gpu ids for gpus that are >95% free.

    Tensorflow does not support checking free memory on gpus.
    This is a crude method that relies on `nvidia-smi` to
    determine which gpus are occupied and which are free.

    Args:
        num_gpus: Number of requested gpus. If not specified,
            ids of all available gpu(s) are returned.

    Returns:
        List[int]: List of gpu ids that are free. Length
            will equal `num_gpus`, if specified.
    """
    # Built-in tensorflow gpu id.
    assert isinstance(num_gpus, (type(None), int))
    if num_gpus == 0:
        return [-1]

    num_requested_gpus = num_gpus
    num_gpus = (
        len(subprocess.check_output("nvidia-smi --list-gpus", shell=True).decode().split("\n")) - 1
    )

    out_str = subprocess.check_output("nvidia-smi | grep MiB", shell=True).decode()
    mem_str = [x for x in out_str.split() if "MiB" in x]
    # First 2 * num_gpu elements correspond to memory for gpus
    # Order: (occupied-0, total-0, occupied-1, total-1, ...)
    mems = [float(x[:-3]) for x in mem_str]
    gpu_percent_occupied_mem = [
        mems[2 * gpu_id] / mems[2 * gpu_id + 1] for gpu_id in range(num_gpus)
    ]

    available_gpus = [gpu_id for gpu_id, mem in enumerate(gpu_percent_occupied_mem) if mem < 0.05]
    if num_requested_gpus and num_requested_gpus > len(available_gpus):
        raise ValueError(
            "Requested {} gpus, only {} are free".format(num_requested_gpus, len(available_gpus))
        )

    return available_gpus[:num_requested_gpus] if num_requested_gpus else available_gpus


def get_world_size():
    """Returns number of gpus currently being used by this process"""
    gpu_ids = os.environ["CUDA_VISIBLE_DEVICES"].split(",")
    if len(gpu_ids) == 1 and gpu_ids[0] == "-1":
        return 0
    return len(gpu_ids)


def package_available(name: str):
    """Returns if package is available.

    Args:
        name (str): Name of the package.

    Returns:
        bool: Whether module exists in environment.
    """
    global _SUPPORTED_PACKAGES
    if name not in _SUPPORTED_PACKAGES:
        _SUPPORTED_PACKAGES[name] = importlib.util.find_spec(name) is not None
    return _SUPPORTED_PACKAGES[name]


def get_package_version(package_or_name) -> str:
    """Returns package version.

    Args:
        package_or_name (``module`` or ``str``): Module or name of module.
            This package must have the version accessible through ``<module>.__version__``.

    Returns:
        str: The package version.

    Examples:
        >>> get_version("numpy")
        "1.20.0"
    """
    if isinstance(package_or_name, str):
        if not package_available(package_or_name):
            raise ValueError(f"Package {package_or_name} not available")
        spec = util.find_spec(package_or_name)
        package_or_name = util.module_from_spec(spec)
        spec.loader.exec_module(package_or_name)
    version = package_or_name.__version__
    return version


def is_package_installed(pkg_str) -> bool:
    """Verify that a package dependency is installed and in the expected version range.

    This is useful for optional third-party dependencies where implementation
    changes are not backwards-compatible.

    Args:
        pkg_str (str): The pip formatted dependency string.
            E.g. "numpy", "numpy>=1.0.0", "numpy>=1.0.0,<=1.10.0", "numpy==1.10.0"

    Returns:
        bool: Whether dependency is satisfied.

    Note:
        This cannot resolve packages where the pip name does not match the python
        package name. ``'-'`` characters are automatically changed to ``'_'``.
    """
    ops = {
        "==": lambda x, y: x == y,
        "<=": lambda x, y: x <= y,
        ">=": lambda x, y: x >= y,
        "<": lambda x, y: x < y,
        ">": lambda x, y: x > y,
    }
    comparison_patterns = "(==|<=|>=|>|<)"

    pkg_str = pkg_str.strip()
    pkg_str = pkg_str.replace("-", "_")
    dependency = list(re.finditer(comparison_patterns, pkg_str))

    if len(dependency) == 0:
        return package_available(pkg_str)

    pkg_name = pkg_str[: dependency[0].start()]
    if not package_available(pkg_name):
        return False

    pkg_version = version.Version(get_package_version(pkg_name))
    version_limits = pkg_str[dependency[0].start() :].split(",")

    for vlimit in version_limits:
        comp_loc = list(re.finditer(comparison_patterns, vlimit))
        if len(comp_loc) != 1:
            raise ValueError(f"Invalid version string: {pkg_str}")
        comp_op = vlimit[comp_loc[0].start() : comp_loc[0].end()]
        comp_version = version.Version(vlimit[comp_loc[0].end() :])
        if not ops[comp_op](pkg_version, comp_version):
            return False
    return True


def is_debug() -> bool:
    return os.environ.get("MEDDLR_DEBUG", "") == "True"


def is_pt_lightning() -> bool:
    return os.environ.get("MEDDLR_PT_LIGHTNING", "") == "True"


def is_repro() -> bool:
    return os.environ.get("MEDDLR_REPRO", "") == "True"


def is_profiling_enabled() -> bool:
    return os.environ.get("MEDDLR_PROFILE", "True") == "True"


def profile_memory() -> bool:
    return os.environ.get("MEDDLR_MPROFILE", "") == "True"


def supports_wandb():
    return "wandb" in sys.modules and not is_debug()


def supports_d2() -> bool:
    """Supports detectron2."""
    return "detectron2" in sys.modules


def supports_cupy():
    if "cupy" not in _SUPPORTED_PACKAGES:
        try:
            import cupy  # noqa
        except ImportError:
            _SUPPORTED_PACKAGES["cupy"] = False
    return package_available("cupy")


def pt_version(dtype=int) -> List:
    version = list(_PT_VERSION.split("."))
    version[-1] = version[-1].split("+")[0]
    if not issubclass(dtype, str):
        version = [dtype(x) for x in version]
    return version


def supports_cplx_tensor() -> bool:
    """Returns `True` if complex tensors supported.

    This can be controlled by the environment variable
    "MEDDLR_ENABLE_CPLX_TENSORS", which should be set to
    "True", "False", or "auto". Defaults to "auto" functionality,
    which enables complex tensors if PyTorch>=1.7.0.
    While complex tensors were introduced in PyTorch 1.6, there were
    known bugs. Enhanced complex tensor support offered in PyTorch >=1.7.0.

    Returns:
        bool: `True` if complex tensors are supported.
    """
    env_var = os.environ.get("MEDDLR_ENABLE_CPLX_TENSORS", "auto")
    is_min_version = pt_version() >= [1, 6]
    is_auto_version = pt_version() >= [1, 7]

    if env_var == "auto":
        env_var = str(is_auto_version)

    if env_var.lower() == "false":
        return False
    elif env_var.lower() == "true":
        if not is_min_version:
            raise RuntimeError(f"Complex tensors not supported in PyTorch {_PT_VERSION}")
        if not is_auto_version:
            warnings.warn("Complex tensor support has known breaking bugs for PyTorch <1.7")
        return True
    else:
        raise ValueError(f"Unknown environment value: {env_var}")


def is_main_process():
    py_version = tuple(sys.version_info[0:2])
    return (py_version < (3, 8) and mp.current_process().name == "MainProcess") or (
        py_version >= (3, 8) and mp.parent_process() is None
    )


def settings_dir():
    return os.environ.get("MEDDLR_SETTINGS", _SETTINGS_DIR)


def get_path_manager(key="meddlr") -> PathManager:
    return PathManagerFactory.get(key)


def get_github_url() -> str:
    return _GITHUB_URL



from typing import Sequence

import torch
from torch import nn

import meddlr.ops as oF
from meddlr.ops import complex as cplx
from meddlr.utils import env
from meddlr.utils.deprecated import deprecated

if env.pt_version() >= [1, 6]:
    import torch.fft


@deprecated(vremove="0.1.0", replacement="forward.SenseModel")
class SenseModel(nn.Module):
    """
    A module that computes forward and adjoint SENSE operations.

    The forward operation converts a complex image -> multi-coil kspace.
    The adjoint operation converts multi-coil kspace -> a complex image.

    This module also supports multiple sensitivity maps. This is useful if
    you would like to generate images from multiple estimated sensitivity maps.
    This module also works with single coil inputs as long as the #coils dimension
    is set to 1.

    Attributes:
        maps (torch.Tensor): Sensitivity maps. Shape ``(B, H, W, #coils, #maps, [2])``.
        weights (torch.Tensor, optional): Undersampling masks (if applicable).
            Shape ``(B, H, W)`` or ``(B, H, W, #coils, #coils)``.
    """

    def __init__(self, maps: torch.Tensor, weights: torch.Tensor = None):
        """
        Args:
            maps (torch.Tensor): Sensitivity maps.
            weights (torch.Tensor): Undersampling masks.
                If ``None``, it is assumed that inputs are fully-sampled.
        """
        super().__init__()

        self.maps = maps
        if weights is None:
            self.weights = 1.0
        else:
            self.weights = weights

    def _adjoint_op(self, kspace):
        """
        Args:
            kspace: Shape (B,H,W,#coils,[2])
        Returns:
            image: Shape (B,H,W,#maps,[2])
        """
        image = ifft2(self.weights * kspace)
        if cplx.is_complex_as_real(kspace):
            image = cplx.mul(image.unsqueeze(-2), cplx.conj(self.maps))  # [B,...,#coils,#maps,2]
            return image.sum(-3)
        else:
            image = cplx.mul(image.unsqueeze(-1), cplx.conj(self.maps))  # [B,...,#coils,#maps,1]
            return image.sum(-2)

    def _forward_op(self, image):
        """
        Args:
            image: Shape (B,H,W,#maps,[2])
        Returns:
            kspace: Shape (B,H,W,#coils,[2])
        """
        if cplx.is_complex_as_real(image):
            kspace = cplx.mul(image.unsqueeze(-3), self.maps)  # [B,...,1,#maps,2]
            kspace = self.weights * fft2(kspace.sum(-2))  # [B,...,#coils,2]
        else:
            kspace = cplx.mul(image.unsqueeze(-2), self.maps)
            kspace = self.weights * fft2(kspace.sum(-1))
        return kspace

    def forward(self, input: torch.Tensor, adjoint: bool = False):
        """Run forward or adjoint SENSE operation on the input.

        Depending on if ``adjoint=True``, the input should either be the
        k-space or the complex image. The shapes for these are as follows:
            - kspace: ``(B, H, W, #coils, [2])
            - image: ``(B, H, W, #maps, [2])``

        Args:
            input (torch.Tensor): If ``adjoint=True``, this is the multi-coil k-space,
                else it is the image.
            adjoint (bool, optional): If ``True``, use adjoint operation.

        Returns:
            torch.Tensor: If ``adjoint=True``, the image, else multi-coil k-space.
        """
        if adjoint:
            output = self._adjoint_op(input)
        else:
            output = self._forward_op(input)
        return output


@deprecated(vremove="0.1.0", replacement="ops.fft2c")
def fft2(data):
    """
    Apply centered 2 dimensional Fast Fourier Transform.
    Args:
        data (torch.Tensor): Complex valued input data with the last dimension
            containing real and imaginary components.
        dims (2-tuple): Containing spatial dimension indices.
    Returns:
        torch.Tensor: The FFT of the input.
    """

    assert data.size(-1) == 2 or env.supports_cplx_tensor()
    if data.size(-1) != 2:
        # Complex tensors supported
        assert env.supports_cplx_tensor(), torch.__version__  # torch.__version__ >= 1.7
        ndims = len(list(data.size()))
        dims = (1, 2)

        data = ifftshift(data, dim=dims)
        data = torch.fft.fftn(data, dim=dims, norm="ortho")
        data = fftshift(data, dim=dims)
        return data

    ndims = len(list(data.size()))

    if ndims == 5:
        data = data.permute(0, 3, 1, 2, 4)
    elif ndims == 6:
        data = data.permute(0, 3, 4, 1, 2, 5)
    else:
        raise ValueError("fft2: ndims > 6 not supported!")

    data = ifftshift(data, dim=(-3, -2))
    data = torch.Tensor.fft(data, 2, normalized=True)
    data = fftshift(data, dim=(-3, -2))

    if ndims == 5:
        data = data.permute(0, 2, 3, 1, 4)
    elif ndims == 6:
        data = data.permute(0, 3, 4, 1, 2, 5)
    else:
        raise ValueError("fft2: ndims > 6 not supported!")

    return data


@deprecated(vremove="0.1.0", replacement="ops.ifft2c")
def ifft2(data):
    """
    Apply centered 2-dimensional Inverse Fast Fourier Transform.
    Args:
        data (torch.Tensor): Complex valued input data with the last dimension
            containing real and imaginary components.
        dims (2-tuple): Containing spatial dimension indices.
    Returns:
        torch.Tensor: The IFFT of the input.
    """
    assert data.size(-1) == 2 or env.supports_cplx_tensor()
    if data.size(-1) != 2:
        # Complex tensors supported
        assert env.supports_cplx_tensor(), torch.__version__  # torch.__version__ >= 1.7
        ndims = len(list(data.size()))
        dims = (1, 2)

        data = ifftshift(data, dim=dims)
        data = torch.fft.ifftn(data, dim=dims, norm="ortho")
        data = fftshift(data, dim=dims)
        return data

    ndims = len(list(data.size()))

    if ndims == 5:
        data = data.permute(0, 3, 1, 2, 4)
    elif ndims == 6:
        data = data.permute(0, 3, 4, 1, 2, 5)
    else:
        raise ValueError("ifft2: ndims > 6 not supported!")

    data = ifftshift(data, dim=(-3, -2))
    data = torch.ifft(data, 2, normalized=True)
    data = fftshift(data, dim=(-3, -2))

    if ndims == 5:
        data = data.permute(0, 2, 3, 1, 4)
    elif ndims == 6:
        data = data.permute(0, 3, 4, 1, 2, 5)
    else:
        raise ValueError("ifft2: ndims > 6 not supported!")

    return data


@deprecated(vremove="0.1.0", replacement="ops.complex.rss")
def root_sum_of_squares(x, dim=0):
    """
    Compute the root sum-of-squares (RSS) transform along a given dimension of
    a complex-valued tensor.
    """
    assert x.size(-1) == 2
    return torch.sqrt((x ** 2).sum(dim=-1).sum(dim))


@deprecated(vremove="0.1.0", replacement="ops.time_average")
def time_average(data, dim, eps=1e-6, keepdim=True):
    """
    Computes time average across a specified axis.
    """
    return oF.time_average(data, dim, eps=eps, keepdim=keepdim)


@deprecated(vremove="0.1.0", replacement="ops.sliding_window")
def sliding_window(data, dim, window_size):
    """
    Computes sliding window with circular boundary conditions across a specified
    axis.
    """
    return oF.sliding_window(data, dim, window_size)


@deprecated(vremove="0.1.0", replacement="ops.center_crop")
def center_crop(data, shape):
    """
    Apply a center crop to a batch of images.
    Args:
        data (torch.Tensor): The input tensor to be center cropped.
        shape (list of ints): The output shape. If shape[dim] = -1, then no crop
            will be applied in that dimension.
    """
    return oF.center_crop(data, shape)


@deprecated(vremove="0.1.0", replacement="ops.complex.complex_center_crop_2d")
def complex_center_crop_2d(data, shape):
    """
    Apply a center crop to the input image or batch of complex images.
    Args:
        data (torch.Tensor): The complex input tensor to be center cropped. It should
            have at least 3 dimensions and the cropping is applied along dimensions
            -3 and -2 and the last dimensions should have a size of 2.
        shape (int, int): The output shape. The shape should be smaller than the
            corresponding dimensions of data.
    Returns:
        torch.Tensor: The center cropped image
    """
    assert 0 < shape[0] <= data.shape[-3]
    assert 0 < shape[1] <= data.shape[-2]
    w_from = (data.shape[-3] - shape[0]) // 2
    h_from = (data.shape[-2] - shape[1]) // 2
    w_to = w_from + shape[0]
    h_to = h_from + shape[1]
    return data[..., w_from:w_to, h_from:h_to, :]


@deprecated(vremove="0.1.0", replacement="ops.normalize")
def normalize(data, mean, stddev, eps=0.0):
    """
    Normalize the given tensor using:
        (data - mean) / (stddev + eps)
    Args:
        data (torch.Tensor): Input data to be normalized
        mean (float): Mean value
        stddev (float): Standard deviation
        eps (float): Added to stddev to prevent dividing by zero
    Returns:
        torch.Tensor: Normalized tensor
    """
    return (data - mean) / (stddev + eps)


@deprecated(vremove="0.1.0", replacement="ops.normalize_instance")
def normalize_instance(data, eps=0.0):
    """
    Normalize the given tensor using:
        (data - mean) / (stddev + eps)
    where mean and stddev are computed from the data itself.
    Args:
        data (torch.Tensor): Input data to be normalized
        eps (float): Added to stddev to prevent dividing by zero
    Returns:
        torch.Tensor: Normalized tensor
    """
    mean = data.mean()
    std = data.std()
    return normalize(data, mean, std, eps), mean, std


# Helper functions


@deprecated(vremove="0.1.0", replacement="ops.complex.center_crop")
def roll(x, shift, dim):
    """
    Similar to np.roll but applies to PyTorch Tensors
    """
    return oF.roll(x, shift, dim)


@deprecated(vremove="0.1.0", replacement="ops.fftshift")
def fftshift(x, dim=None):
    """
    Similar to np.fft.fftshift but applies to PyTorch Tensors
    """
    return oF.fftshift(x, dim=dim)


@deprecated(vremove="0.1.0", replacement="ops.ifftshift")
def ifftshift(x, dim=None):
    """
    Similar to np.fft.ifftshift but applies to PyTorch Tensors
    """
    return oF.ifftshift(x, dim=dim)


@deprecated(vremove="0.1.0", replacement="ops.pad")
def pad(x: torch.Tensor, shape: Sequence[int], mode="constant", value=0):
    """
    Args:
        x: Input tensor of shape (B, ...)
        shape: Shape to zero pad to. Use `None` to skip padding certain dimensions.
    Returns:
    """
    return oF.pad(x, shape, mode=mode, value=value)


@deprecated(vremove="0.1.0", replacement="ops.zero_pad")
def zero_pad(x: torch.Tensor, shape: Sequence[int]):
    return oF.zero_pad(x, shape)



"""
Adapted from
https://github.com/facebookresearch/detectron2
"""
import importlib
import os
import re
import subprocess
import sys
from collections import defaultdict

import numpy as np
import PIL
import torch
import torchvision
from tabulate import tabulate

__all__ = ["collect_env_info"]


def collect_torch_env():
    try:
        import torch.__config__

        return torch.__config__.show()
    except ImportError:
        # compatible with older versions of pytorch
        from torch.utils.collect_env import get_pretty_env_info

        return get_pretty_env_info()


def detect_compute_compatibility(CUDA_HOME, so_file):
    try:
        cuobjdump = os.path.join(CUDA_HOME, "bin", "cuobjdump")
        if os.path.isfile(cuobjdump):
            output = subprocess.check_output(
                "'{}' --list-elf '{}'".format(cuobjdump, so_file), shell=True
            )
            output = output.decode("utf-8").strip().split("\n")
            sm = []
            for line in output:
                line = re.findall(r"\.sm_[0-9]*\.", line)[0]
                sm.append(line.strip("."))
            sm = sorted(set(sm))
            return ", ".join(sm)
        else:
            return so_file + "; cannot find cuobjdump"
    except Exception:
        # unhandled failure
        return so_file


def collect_env_info():
    has_cuda = torch.cuda.is_available()

    data = []
    data.append(("sys.platform", sys.platform))
    data.append(("Python", sys.version.replace("\n", "")))
    data.append(("numpy", np.__version__))

    data.append(("PyTorch", torch.__version__ + " @" + os.path.dirname(torch.__file__)))
    data.append(("PyTorch debug build", torch.version.debug))

    data.append(("CUDA available", has_cuda))
    if has_cuda:
        devices = defaultdict(list)
        for k in range(torch.cuda.device_count()):
            devices[torch.cuda.get_device_name(k)].append(str(k))
        for name, devids in devices.items():
            data.append(("GPU " + ",".join(devids), name))

        # NOTE: the use of CUDA_HOME requires the CUDA build deps, though in
        # theory meddlr should be made runnable with only the CUDA runtime
        from torch.utils.cpp_extension import CUDA_HOME

        data.append(("CUDA_HOME", str(CUDA_HOME)))

        if CUDA_HOME is not None and os.path.isdir(CUDA_HOME):
            try:
                nvcc = os.path.join(CUDA_HOME, "bin", "nvcc")
                nvcc = subprocess.check_output("'{}' -V | tail -n1".format(nvcc), shell=True)
                nvcc = nvcc.decode("utf-8").strip()
            except subprocess.SubprocessError:
                nvcc = "Not Available"
            data.append(("NVCC", nvcc))

        cuda_arch_list = os.environ.get("TORCH_CUDA_ARCH_LIST", None)
        if cuda_arch_list:
            data.append(("TORCH_CUDA_ARCH_LIST", cuda_arch_list))
    data.append(("Pillow", PIL.__version__))

    try:
        data.append(
            (
                "torchvision",
                str(torchvision.__version__) + " @" + os.path.dirname(torchvision.__file__),
            )
        )
        if has_cuda:
            try:
                torchvision_C = importlib.util.find_spec("torchvision._C").origin
                msg = detect_compute_compatibility(CUDA_HOME, torchvision_C)
                data.append(("torchvision arch flags", msg))
            except ImportError:
                data.append(("torchvision._C", "failed to find"))
    except AttributeError:
        data.append(("torchvision", "unknown"))

    # Slurm info
    data.append(("SLURM_JOB_ID", os.environ.get("SLURM_JOB_ID", "slurm not detected")))

    try:
        import cv2

        data.append(("cv2", cv2.__version__))
    except ImportError:
        pass
    env_str = tabulate(data) + "\n"
    env_str += collect_torch_env()
    return env_str


if __name__ == "__main__":
    print(collect_env_info())



from typing import Sequence, Tuple, Union

import torch

from meddlr.transforms.base.mask import KspaceMaskTransform
from meddlr.transforms.build import TRANSFORM_REGISTRY
from meddlr.transforms.transform import NoOpTransform
from meddlr.transforms.transform_gen import TransformGen

__all__ = ["RandomKspaceMask"]


@TRANSFORM_REGISTRY.register()
class RandomKspaceMask(TransformGen):
    """A model that generates random masks."""

    _base_transform = KspaceMaskTransform
    _param_names = ["rhos"]

    def __init__(
        self,
        p: float = 0.0,
        rhos: Union[float, Tuple[float, float]] = None,
        kind: str = "uniform",
        std_scale: float = None,
        per_example=False,
        calib_size=None,
    ):
        if isinstance(rhos, Sequence) and len(rhos) == 1:
            rhos = rhos[0]

        params = locals()
        params = {k: params[k] for k in list(self._param_names) + ["p"]}
        super().__init__(params=params, p=p)

        self.kind = kind
        self.per_example = per_example
        self.calib_size = calib_size
        self.std_scale = std_scale

    def get_transform(self, input: torch.Tensor):
        params = self._get_param_values(use_schedulers=True)
        rhos = params["rhos"]

        if self._rand() >= params["p"]:
            return NoOpTransform()

        if isinstance(rhos, Sequence):
            rho = self._rand_range(*rhos)
        else:
            rho = rhos

        gen = self._generator
        if gen is None or gen.device != input.device:
            gen = torch.Generator(device=input.device).manual_seed(int(self._rand() * 1e10))
        return KspaceMaskTransform(
            rho=rho,
            kind=self.kind,
            per_example=self.per_example,
            calib_size=self.calib_size,
            generator=gen,
            std_scale=self.std_scale,
        )



from torch.utils.data import DistributedSampler

from meddlr.data.samplers.group_sampler import (
    AlternatingGroupSampler,
    DistributedGroupSampler,
    GroupSampler,
)
from meddlr.data.samplers.sampler import AlternatingSampler

__all__ = ["build_train_sampler", "build_val_sampler"]


def build_train_sampler(cfg, dataset, distributed=False):
    """Builds the training sampler from the config.

    Args:
        cfg (CfgNode): The config. The ``cfg.DATALOADER.SAMPLER_TRAIN``field
            will be used to determine the sampler type.
        dataset (torch.data.Dataset): The training dataset. This dataset
            must follow the Meddlr dataset convention. See :cls:`SliceDataset`
            for more information.
        distributed (bool, optional): Whether to use a distributed sampler.
            Note custom samplers (e.g. AlternatingSampler/GroupSampler)
            do not support this argument.

    Returns:
        torch.utils.data.sampler.Sampler: The sampler.
    """
    sampler = cfg.DATALOADER.SAMPLER_TRAIN
    is_batch_sampler = False
    seed = cfg.SEED if cfg.SEED > -1 else None
    if sampler == "AlternatingSampler":
        sampler = AlternatingSampler(
            dataset,
            T_s=cfg.DATALOADER.ALT_SAMPLER.PERIOD_SUPERVISED,
            T_us=cfg.DATALOADER.ALT_SAMPLER.PERIOD_UNSUPERVISED,
            seed=seed,
        )
    elif sampler == "GroupSampler":
        is_batch_sampler = cfg.DATALOADER.GROUP_SAMPLER.AS_BATCH_SAMPLER
        sampler = GroupSampler(
            dataset,
            batch_by=cfg.DATALOADER.GROUP_SAMPLER.BATCH_BY,
            batch_size=cfg.SOLVER.TRAIN_BATCH_SIZE,
            as_batch_sampler=is_batch_sampler,
            drop_last=cfg.DATALOADER.DROP_LAST,
            shuffle=True,
            seed=seed,
        )
    elif sampler == "AlternatingGroupSampler":
        is_batch_sampler = cfg.DATALOADER.GROUP_SAMPLER.AS_BATCH_SAMPLER
        sampler = AlternatingGroupSampler(
            dataset,
            T_s=cfg.DATALOADER.ALT_SAMPLER.PERIOD_SUPERVISED,
            T_us=cfg.DATALOADER.ALT_SAMPLER.PERIOD_UNSUPERVISED,
            batch_by=cfg.DATALOADER.GROUP_SAMPLER.BATCH_BY,
            batch_size=cfg.SOLVER.TRAIN_BATCH_SIZE,
            as_batch_sampler=is_batch_sampler,
            drop_last=cfg.DATALOADER.DROP_LAST,
            seed=seed,
        )
    elif sampler == "":
        sampler = DistributedSampler(dataset, shuffle=True) if distributed else None
    else:
        raise ValueError("Unknown Sampler {}".format(sampler))

    return sampler, is_batch_sampler


def build_val_sampler(cfg, dataset, distributed: bool = False, dist_group_by="file_name"):
    """
    Build the whole model architecture, defined by ``cfg.MODEL.META_ARCHITECTURE``.
    Note that it does not load any weights from ``cfg``.
    """
    sampler = cfg.DATALOADER.SAMPLER_TRAIN
    is_batch_sampler = False
    seed = cfg.SEED if cfg.SEED > -1 else None
    if (
        sampler in ("GroupSampler", "AlternatingGroupSampler")
        and cfg.DATALOADER.GROUP_SAMPLER.BATCH_BY
    ):
        is_batch_sampler = True
        sampler = GroupSampler(
            dataset,
            batch_by=cfg.DATALOADER.GROUP_SAMPLER.BATCH_BY,
            batch_size=cfg.SOLVER.TEST_BATCH_SIZE,
            as_batch_sampler=is_batch_sampler,
            drop_last=False,
            shuffle=False,
            seed=seed,
        )
    elif sampler == "":
        sampler = (
            DistributedGroupSampler(dataset, group_by=dist_group_by, shuffle=False)
            if distributed
            else None
        )
    else:
        sampler = None

    return sampler, is_batch_sampler



import logging
from enum import Enum
from typing import Callable, Sequence, Union

import numpy as np
import torch

logger = logging.getLogger(__name__)

arr_type = Union[np.ndarray, torch.Tensor]


def to_bool(x):
    """Converts a tensor or ndarray into the bool dtype.

    This function first checks if the tensory is already of
    the bool dtype prior to doing conversion.
    NumPy (and potentially torch) has some issues with doing
    this check first for shared-memory tensors.

    Args:
        x (array-like): An array-like or torch Tensor to convert to bool dtype.

    Returns:
        np.ndarray | torch.Tensor: The array with dtype ``bool``.
    """
    if isinstance(x, torch.Tensor):
        x = x.type(torch.bool) if x.dtype != torch.bool else x
    else:
        if not isinstance(x, np.ndarray):
            x = np.asarray(x, dtype=np.bool)
        x = x.astype(np.bool) if x.dtype != np.bool else x
    return x


def flatten_other_dims(
    xs: Union[arr_type, Sequence[arr_type]], dim: Union[int, Sequence[int]] = None
):
    single_input = isinstance(xs, (np.ndarray, torch.Tensor))
    is_tensor = isinstance(xs, torch.Tensor) if single_input else isinstance(xs[0], torch.Tensor)

    if single_input:
        xs = [xs]
    dim = (dim,) if isinstance(dim, int) else tuple(dim)

    if dim:
        shape = tuple(xs[0].shape[d] for d in dim) + (-1,)
        dims_ordered = dim + tuple(i for i in range(xs[0].ndim) if i not in dim)
        if is_tensor:
            xs = [x.permute(dims_ordered) for x in xs]
        else:
            xs = [x.transpose(dims_ordered) for x in xs]
        xs = (x.reshape(shape) for x in xs)
    else:
        xs = (x.flatten() for x in xs)

    if single_input:
        xs = list(xs)[0]
    return xs


def flatten_non_category_dims(xs: Union[arr_type, Sequence[arr_type]], category_dim: int = None):
    """Flattens all non-category dimensions into a single dimension.

    Args:
        xs (ndarrays): Sequence of ndarrays with the same category dimension.
        category_dim: The dimension/axis corresponding to different categories.
            i.e. `C`. If `None`, behaves like `np.flatten(x)`.

    Returns:
        ndarray: Shape (C, -1) if `category_dim` specified else shape (-1,)
    """
    return flatten_other_dims(xs, dim=category_dim)


def rms_cv(y_pred: arr_type, y_true: arr_type, dim=None):
    """Compute root-mean-squared coefficient of variation.

    This is typically done to compare intra-method variability.
    For example if multiple measurements are taken using the same method.
    However, in many segmentation manuscripts, this is equation is also
    used.

    This quantity is symmetric.

    Args:
        y_pred (ndarray): Measurements from trial 1.
        y_true (ndarray): Measurements from trial 2.
        dim (int, optional): Dimension/axis over which to compute metric.
            If `None`, all dimensions will be reduced.

    Returns:
        ndarray: If `dim=None`, scalar value.
    """
    is_tensor = isinstance(y_pred, torch.Tensor)
    if is_tensor:
        cat_tensor = torch.stack([y_pred, y_true], dim=0).type(torch.float32)
        stds = torch.std(cat_tensor, dim=0)
        means = torch.mean(cat_tensor, dim=0)
        cv = stds / means
        return torch.sqrt(torch.mean(cv ** 2, dim=dim))
    else:
        stds = np.std([y_pred, y_true], axis=0)
        means = np.mean([y_pred, y_true], axis=0)
        cv = stds / means
        return np.sqrt(np.mean(cv ** 2, axis=dim))


def rmse_cv(y_pred: np.ndarray, y_true: np.ndarray, dim=None):
    """Compute root-mean-squared error coefficient of variation.

    This quantity is not symmetric.

    Args:
        y_pred (ndarray): Predicted measurements.
        y_true (ndarray): Ground-truth/baseline measurements.
        dim (int, optional): Dimension/axis over which to compute metric.
            If `None`, all dimensions will be reduced.

    Returns:
        ndarray: If `dim=None`, scalar value.
    """
    is_tensor = isinstance(y_pred, torch.Tensor)
    if is_tensor:
        rmse = torch.sqrt(torch.mean((y_pred - y_true) ** 2, dim=dim))
        means = torch.abs(torch.mean(y_true, dim=dim))
    else:
        rmse = np.sqrt(np.mean((y_pred - y_true) ** 2, axis=dim))
        means = np.absolute(np.mean(y_true, axis=dim))
    return rmse / means


class Reductions(Enum):
    RMS_CV = 1, "RMS-CV", rms_cv
    RMSE_CV = 2, "RMSE-CV", rmse_cv

    def __new__(cls, value: int, display_name: str, func: Callable):
        """
        Args:
            value (int): Unique integer value.
            patterns (`List[str]`): List of regex patterns that would match the
                hostname on the compute cluster. There can be multiple hostnames
                per compute cluster because of the different nodes.
            save_dir (str): Directory to save data to.
        """
        obj = object.__new__(cls)
        obj._value_ = value

        obj.display_name = display_name
        obj.func = func

        return obj



from meddlr.modeling.layers import build, conv, gauss
from meddlr.modeling.layers.build import CUSTOM_LAYERS_REGISTRY  # noqa: F401
from meddlr.modeling.layers.conv import ConvWS2d, ConvWS3d  # noqa: F401
from meddlr.modeling.layers.gauss import GaussianBlur  # noqa: F401

__all__ = []
__all__.extend(build.__all__)
__all__.extend(conv.__all__)
__all__.extend(gauss.__all__)



from typing import Sequence

import numpy as np
import torch

from meddlr.ops import complex as cplx
from meddlr.ops.utils import roll
from meddlr.utils import env

if env.pt_version() >= [1, 6]:
    import torch.fft

__all__ = [
    "fftnc",
    "ifftnc",
    "fftc",
    "ifftc",
    "fft2c",
    "ifft2c",
    "fft3c",
    "ifft3c",
    "fftshift",
    "ifftshift",
]


def fftnc(input: torch.Tensor, dim=None, norm="ortho", is_real: bool = None) -> torch.Tensor:
    """Apply nD centered fast fourier transform.

    This function is a backwards-compatible wrapper for centered :meth:`torch.fft.fftn`.
    It supports backwards compatibility with ``torch.fft``as implemented in torch<1.7.

    Args:
        input (torch.Tensor): A tensor (typically complex).
        dim (Tuple[int]): Dimensions to be transformed.
        norm (str | bool, optional): The normalization method.
            Defaults to ``'ortho'``. For torch<1.7, only ``'ortho'``
            is supported.
        is_real (bool, optional): If ``True``, ``input`` is a real-valued
            tensor. If ``None`` or ``False`` and ``input.shape[-1] == 2``,
            ``input`` is a real-view of a complex tensor.

    Returns:
        torch.Tensor

    Note:
        Real-valued tensors are not supported with ``torch<1.7``.
    """
    return _fft_template(input, kind="fft", dim=dim, norm=norm, is_real=is_real)


def ifftnc(input: torch.Tensor, dim=None, norm="ortho", is_real: bool = None) -> torch.Tensor:
    """Apply nD centered inverse fast fourier transform.

    This supports backwards compatibility with ``torch.fft``
    as implemented in torch<1.7.

    Args:
        input (torch.Tensor): A tensor (typically complex).
        dim (Tuple[int]): Dimensions to be transformed.
        norm (str | bool, optional): The normalization method.
            Defaults to ``'ortho'``. For torch<1.7, only ``'ortho'``
            is supported.
        is_real (bool, optional): If ``True``, ``input`` is a real-valued
            tensor. If ``None`` or ``False`` and ``input.shape[-1] == 2``,
            ``input`` is a real-view of a complex tensor.

    Returns:
        torch.Tensor

    Note:
        Real-valued tensors are not supported with ``torch<1.7``.
    """
    return _fft_template(input, kind="ifft", dim=dim, norm=norm, is_real=is_real)


def fftc(input: torch.Tensor, norm: str = "ortho", is_real: bool = None, channels_last=False):
    """Apply 1D centered Fast Fourier Transform (FFT).

    Args:
        input (torch.Tensor): A tensor.
        norm (str | bool, optional): The normalization method.
            Defaults to ``'ortho'``. For torch<1.7, only ``'ortho'``
            is supported.
        is_real (bool, optional): If ``True``, ``input`` is a real-valued
            tensor. If ``None`` or ``False`` and ``input.shape[-1] == 2``,
            ``input`` is a real-view of a complex tensor.
        channels_last (bool, optional): If ``True``, apply to first
            non-batch dimensions. If ``False``, apply to last dimension.

    Returns:
        torch.Tensor: The 1D centered FFT of the input.
    """
    dim = _get_fft_dims(input, 1, is_real=is_real, channels_last=channels_last)
    return fftnc(input, dim=dim, norm=norm, is_real=is_real)


def ifftc(input: torch.Tensor, channels_last=False, norm: str = "ortho", is_real: bool = None):
    """Apply centered 2 dimensional Fast Fourier Transform.

    Args:
        data (torch.Tensor): Complex valued input data with the last dimension
            containing real and imaginary components.
        dims (2-tuple): Containing spatial dimension indices.
    Returns:
        torch.Tensor: The FFT of the input.
    """
    dim = _get_fft_dims(input, 1, is_real=is_real, channels_last=channels_last)
    return ifftnc(input, dim=dim, norm=norm, is_real=is_real)


def fft2c(input: torch.Tensor, channels_last=False, norm: str = "ortho", is_real: bool = None):
    """Apply centered 2 dimensional Fast Fourier Transform.

    Args:
        data (torch.Tensor): Complex valued input data with the last dimension
            containing real and imaginary components.
        dims (2-tuple): Containing spatial dimension indices.
    Returns:
        torch.Tensor: The FFT of the input.
    """
    dim = _get_fft_dims(input, 2, is_real=is_real, channels_last=channels_last)
    return fftnc(input, dim=dim, norm=norm, is_real=is_real)


def ifft2c(input, channels_last=False, norm: str = "ortho", is_real: bool = None):
    """Apply centered 2-dimensional Inverse Fast Fourier Transform.

    Args:
        data (torch.Tensor): Complex valued input data with the last dimension
            containing real and imaginary components.
        dims (2-tuple): Containing spatial dimension indices.
    Returns:
        torch.Tensor: The IFFT of the input.
    """
    dim = _get_fft_dims(input, 2, is_real=is_real, channels_last=channels_last)
    return ifftnc(input, dim=dim, norm=norm, is_real=is_real)


def fft3c(input: torch.Tensor, channels_last=False, norm: str = "ortho", is_real: bool = None):
    """Apply centered 2 dimensional Fast Fourier Transform.

    Args:
        data (torch.Tensor): Complex valued input data with the last dimension
            containing real and imaginary components.
        dims (2-tuple): Containing spatial dimension indices.
    Returns:
        torch.Tensor: The FFT of the input.
    """
    dim = _get_fft_dims(input, 3, is_real=is_real, channels_last=channels_last)
    return fftnc(input, dim=dim, norm=norm, is_real=is_real)


def ifft3c(input, channels_last=False, norm: str = "ortho", is_real: bool = None):
    """Apply centered 2-dimensional Inverse Fast Fourier Transform.

    Args:
        data (torch.Tensor): Complex valued input data with the last dimension
            containing real and imaginary components.
        dims (2-tuple): Containing spatial dimension indices.
    Returns:
        torch.Tensor: The IFFT of the input.
    """
    dim = _get_fft_dims(input, 3, is_real=is_real, channels_last=channels_last)
    return ifftnc(input, dim=dim, norm=norm, is_real=is_real)


def fftshift(x, dim=None):
    """
    Similar to np.fft.fftshift but applies to PyTorch Tensors
    """
    if dim is None:
        dim = tuple(range(x.dim()))
        shift = [dim // 2 for dim in x.shape]
    elif isinstance(dim, int):
        shift = x.shape[dim] // 2
    else:
        shift = [x.shape[i] // 2 for i in dim]
    return roll(x, shift, dim)


def ifftshift(x, dim=None):
    """
    Similar to np.fft.ifftshift but applies to PyTorch Tensors
    """
    if dim is None:
        dim = tuple(range(x.dim()))
        shift = [(dim + 1) // 2 for dim in x.shape]
    elif isinstance(dim, int):
        shift = (x.shape[dim] + 1) // 2
    else:
        shift = [(x.shape[i] + 1) // 2 for i in dim]
    return roll(x, shift, dim)


def _fft_template(
    data: torch.Tensor, kind, dim=None, norm="ortho", is_real: bool = None, centered: bool = True
) -> torch.Tensor:
    """Template for fft operations."""
    if isinstance(dim, int):
        dim = (dim,)
    if norm is True:
        norm = "ortho"
    assert kind in ("fft", "ifft")

    if is_real is None:
        is_real = not (cplx.is_complex_as_real(data) or cplx.is_complex(data))
    ndim = data.ndim

    if not env.supports_cplx_tensor():
        # Defaults to torch.fft method.
        assert norm in ("ortho", False)  # norm not supported
        norm = norm == "ortho"
        assert not is_real  # real tensors not supported
        assert cplx.is_complex_as_real(data)

        dim = tuple(sorted(_to_positive_index(dim)))
        if ndim - 1 in dim:
            raise ValueError("Cannot take fft along the real/imaginary channel.")
        if len(set(dim)) != len(dim):
            raise ValueError(f"Expected unique dimensions, got {dim}.")

        signal_ndim = len(dim)
        if signal_ndim > 3:
            raise ValueError(f"Number of dimensions must be <=3, got {len(dim)}.")

        # Reorder dims (if necessary).
        last_dims = dim + (ndim - 1,)
        permute = last_dims != tuple(range(ndim - signal_ndim - 1, ndim))
        if permute:
            order = tuple(i for i in range(ndim)) + last_dims
            data = data.permute(order)

        shift_dims = tuple([-2 - i for i in range(len(dim))][::-1])

        if kind == "fft":
            if centered:
                data = ifftshift(data, dim=shift_dims)
            data = torch.Tensor.fft(data, signal_ndim, normalized=norm)
            if centered:
                data = fftshift(data, shift_dims)
        elif kind == "ifft":
            if centered:
                data = ifftshift(data, dim=shift_dims)
            data = torch.ifft(data, signal_ndim, normalized=norm)
            if centered:
                data = fftshift(data, shift_dims)
        else:
            raise ValueError(f"Unknown `kind={kind}`")

        # Reorder dims (if necessary).
        if permute:
            reorder = tuple(np.argsort(order))
            data = data.permute(reorder)
        return data

    is_real_view = not is_real and cplx.is_complex_as_real(data)
    if is_real_view:
        # Make dimensions positive relative to input dimensions.
        dim = _to_positive_index(dim, ndim=data.ndim)
        data = torch.view_as_complex(data)

    if kind == "fft":
        if centered:
            data = ifftshift(data, dim=dim)
        data = torch.fft.fftn(data, dim=dim, norm=norm)
        if centered:
            data = fftshift(data, dim=dim)
    elif kind == "ifft":
        if centered:
            data = ifftshift(data, dim=dim)
        data = torch.fft.ifftn(data, dim=dim, norm=norm)
        if centered:
            data = fftshift(data, dim=dim)
    else:
        raise ValueError(f"Unknown `kind={kind}`")
    if is_real_view:
        data = torch.view_as_real(data)
    return data


def _get_fft_dims(x, signal_ndim, is_real, channels_last):
    if channels_last:
        return tuple(range(1, 1 + signal_ndim))

    if not is_real and not cplx.is_complex(x) and cplx.is_complex_as_real(x):
        return tuple(range(-1 - signal_ndim, -1))
    else:
        return tuple(range(-signal_ndim, 0))


def _to_positive_index(idxs: Sequence[int], ndim: int):
    return tuple(ndim + i if i < 0 else i for i in idxs)



import itertools

import torch
import torchvision.utils as tv_utils
from torch import nn

import meddlr.ops.complex as cplx
from meddlr.config.config import configurable
from meddlr.data.transforms.noise import NoiseModel
from meddlr.modeling.meta_arch.build import META_ARCH_REGISTRY, build_model
from meddlr.utils.events import get_event_storage
from meddlr.utils.general import move_to_device
from meddlr.utils.transforms import SenseModel

__all__ = ["DenoisingModel"]


@META_ARCH_REGISTRY.register()
class DenoisingModel(nn.Module):
    @configurable
    def __init__(
        self,
        model,
        noiser: NoiseModel,
        use_fully_sampled_target=False,
        use_fully_sampled_target_eval: bool = None,
        vis_period: int = -1,
    ):
        """
        Args:
            model (nn.Module): The base model.
            noiser (NoiseModel): The additive noise model.
            use_fully_sampled_target (bool, optional): If ``True``,
                use fully sampled images as the target for denoising during training.
            use_fully_sampled_target_eval (bool, optional): If ``True``,
                use fully sampled images as the target for denoising during evaluation
                (includes validation). If ``None``, defaults to ``use_fully_sampled_target``.
        """
        super().__init__()
        self.model = model
        self.noiser = noiser

        # Default validation to use_fully_sampled_target_eval.
        if use_fully_sampled_target_eval is None:
            use_fully_sampled_target_eval = use_fully_sampled_target
        self.use_fully_sampled_target = use_fully_sampled_target
        self.use_fully_sampled_target_eval = use_fully_sampled_target_eval

        # Visualization done by this model
        if hasattr(self.model, "vis_period") and vis_period > 0:
            self.model.vis_period = -1
        self.vis_period = vis_period

    def augment(self, kspace):
        """Noise augmentation module.
        TODO: Perform the augmentation here.
        """
        kspace = kspace.detach().clone()
        return self.noiser(kspace, clone=False)

    def visualize_training(self, kspace, zfs, targets, preds):
        """A function used to visualize reconstructions.

        Args:
            targets: NxHxWx2 tensors of target images.
            preds: NxHxWx2 tensors of predictions.
        """
        storage = get_event_storage()

        with torch.no_grad():
            if cplx.is_complex(kspace):
                kspace = torch.view_as_real(kspace)
            kspace = kspace[0, ..., 0, :].unsqueeze(0).cpu()  # calc mask for first coil only
            targets = targets[0, ...].unsqueeze(0).cpu()
            preds = preds[0, ...].unsqueeze(0).cpu()
            zfs = zfs[0, ...].unsqueeze(0).cpu()

            all_images = torch.cat([zfs, preds, targets], dim=2)

            imgs_to_write = {
                "phases": cplx.angle(all_images),
                "images": cplx.abs(all_images),
                "errors": cplx.abs(preds - targets),
                "masks": cplx.get_mask(kspace),
            }

            for name, data in imgs_to_write.items():
                data = data.squeeze(-1).unsqueeze(1)
                data = tv_utils.make_grid(data, nrow=1, padding=1, normalize=True, scale_each=True)
                storage.put_image("train/{}".format(name), data.numpy(), data_format="CHW")

    def forward(self, inputs, return_pp=False, vis_training=False):
        """
        TODO: condense into list of dataset dicts.
        Args:
            inputs: Standard ss_recon module input dictionary
                * "kspace": Kspace. If fully sampled, and want to simulate
                    undersampled kspace, provide "mask" argument.
                * "maps": Sensitivity maps
                * "target" (optional): Target image (typically fully sampled).
                * "mask" (optional): Undersampling mask to apply.
                * "signal_model" (optional): The signal model. If provided,
                    "maps" will not be used to estimate the signal model.
                    Use with caution.
            return_pp (bool, optional): If `True`, return post-processing
                parameters "mean", "std", and "norm" if included in the input.
            vis_training (bool, optional): If `True`, force visualize training
                on this pass. Can only be `True` if model is in training mode.

        Returns:
            Dict: A standard ss_recon output dict
                * "pred": The reconstructed image
                * "target" (optional): The target image.
                    Added if provided in the input.
                * "mean"/"std"/"norm" (optional): Pre-processing parameters.
                    Added if provided in the input.
                * "zf_image": The zero-filled image.
                    Added when model is in eval mode.
        """
        if vis_training and not self.training:
            raise ValueError("vis_training is only applicable in training mode.")

        device = next(self.parameters()).device
        inputs = move_to_device(inputs, device)

        if self.training and self.vis_period > 0:
            storage = get_event_storage()
            if storage.iter % self.vis_period == 0:
                vis_training = True

        if not any(k in inputs for k in ["supervised", "unsupervised"]):
            inputs = {"supervised": inputs}

        use_fully_sampled_target = (self.training and self.use_fully_sampled_target) or (
            not self.training and self.use_fully_sampled_target_eval
        )

        if "supervised" in inputs:
            sup_inputs = inputs["supervised"]
            if use_fully_sampled_target:
                img = sup_inputs["target"]
                A = SenseModel(sup_inputs["maps"])
                sup_inputs["kspace"] = self.augment(A(img, adjoint=False))
            else:
                kspace = sup_inputs["kspace"]
                A = SenseModel(sup_inputs["maps"], weights=cplx.get_mask(kspace))
                sup_inputs["target"] = A(kspace, adjoint=True).detach()
                sup_inputs["kspace"] = self.augment(kspace)
        if "unsupervised" in inputs:
            unsup_inputs = inputs["unsupervised"]
            kspace = unsup_inputs["kspace"]
            A = SenseModel(unsup_inputs["maps"], weights=cplx.get_mask(kspace))
            unsup_inputs["target"] = A(kspace, adjoint=True).detach()
            unsup_inputs["kspace"] = self.augment(kspace)

        keys = [set(v.keys()) for v in inputs.values()]
        keys = keys[0].intersection(*keys)
        inputs = {k: [inputs[field][k] for field in inputs.keys()] for k in keys}
        inputs = {
            k: torch.cat(inputs[k], dim=0)
            if isinstance(inputs[k][0], torch.Tensor)
            else itertools.chain(inputs[k])
            for k in inputs.keys()
        }

        output_dict = self.model(inputs, return_pp=True, vis_training=vis_training)

        return output_dict

    @classmethod
    def from_config(cls, cfg):
        device = torch.device(cfg.MODEL.DEVICE)

        model_cfg = cfg.clone()
        model_cfg.defrost()
        model_cfg.MODEL.META_ARCHITECTURE = cfg.MODEL.DENOISING.META_ARCHITECTURE
        model_cfg.freeze()
        model = build_model(model_cfg)

        noise_cfg = cfg.clone()
        noise_cfg.defrost()
        noise_cfg.MODEL.CONSISTENCY.AUG.NOISE.STD_DEV = cfg.MODEL.DENOISING.NOISE.STD_DEV
        noise_cfg.freeze()
        noiser = NoiseModel.from_cfg(noise_cfg, device=device)

        use_fully_sampled_target = cfg.MODEL.DENOISING.NOISE.USE_FULLY_SAMPLED_TARGET
        use_fully_sampled_target_eval = cfg.MODEL.DENOISING.NOISE.USE_FULLY_SAMPLED_TARGET_EVAL
        return {
            "model": model,
            "noiser": noiser,
            "use_fully_sampled_target": use_fully_sampled_target,
            "use_fully_sampled_target_eval": use_fully_sampled_target_eval,
        }



"""Basic Transforms.
"""
from functools import partial

import numpy as np
import torch
from fvcore.common.registry import Registry

from meddlr.forward import SenseModel
from meddlr.ops import complex as cplx
from meddlr.utils import transforms as T

from .motion import MotionModel
from .noise import NoiseModel

NORMALIZER_REGISTRY = Registry("NORMALIZER")
NORMALIZER_REGISTRY.__doc__ = """
Registry for normalizing images
"""


def build_normalizer(cfg):
    cfg = cfg.MODEL.NORMALIZER
    name = cfg.NAME
    obj = NORMALIZER_REGISTRY.get(name)(keywords=cfg.KEYWORDS)
    return obj


def normalize_affine(x, bias, scale):
    return (x - bias) / scale


def unnormalize_affine(x, bias, scale):
    return x * scale + bias


class Normalizer:
    """Template for normalizing and undoing normalization for scans."""

    # Keywords of dictionary keys to process (if they exist)
    # image: The zero-filled or reconstructed image
    # target: The target (fully-sampled) image
    # masked_kspace: The kspace used to calculate the zero-filled image.
    KEYWORDS = ("image", "target", "masked_kspace")

    def __init__(self, keywords=None):
        if not keywords:
            keywords = self.KEYWORDS
        # Copy the sequence to allow modification down the line.
        self._keywords = tuple(keywords)

    def normalize(self, **kwargs):
        pass

    def undo(self, **kwargs):
        pass


@NORMALIZER_REGISTRY.register()
class NoOpNormalizer(Normalizer):
    def normalize(self, **kwargs):
        outputs = {k: v for k, v in kwargs.items()}
        outputs.update(
            {
                "mean": torch.tensor([0.0], dtype=torch.float32),
                "std": torch.tensor([1.0], dtype=torch.float32),
            }
        )
        return outputs

    def undo(self, **kwargs):
        return {k: v for k, v in kwargs.items()}


@NORMALIZER_REGISTRY.register()
class AffineNormalizer(Normalizer):
    def normalize(self, scale, bias, **kwargs):
        outputs = {}
        for k, v in kwargs.items():
            if k in self._keywords:
                outputs[k] = (v - bias) / scale
        outputs.update({"mean": bias, "std": scale})
        return outputs

    def undo(self, mean, std, channels_last=False, **kwargs):
        image = kwargs["image"]
        mean = self._reshape_params(mean, image.ndim, channels_last=channels_last).to(image.device)
        std = self._reshape_params(std, image.ndim, channels_last=channels_last).to(image.device)

        outputs = {}
        for kw in ("image", "target"):
            if kw in self._keywords:
                outputs[kw] = unnormalize_affine(kwargs[kw], mean, std)
        if any("kspace" in k for k in kwargs.keys()):
            raise ValueError("Currently does not support undoing analysis on kspace")

        # Add other keys that were not computed.
        outputs.update({k: v for k, v in kwargs.items() if k not in outputs})
        return outputs

    def _reshape_params(self, param, ndim: int, channels_last: bool = False) -> torch.Tensor:
        if not channels_last or param.ndim != 2:
            return param.view(param.shape + (1,) * (ndim - param.ndim))
        shape = (param.shape[0],) + (1,) * (ndim - param.ndim) + (param.shape[1],)
        return param.reshape(shape)


@NORMALIZER_REGISTRY.register()
class TopMagnitudeNormalizer(AffineNormalizer):
    """Normalizes by percentile of magnitude values."""

    def __init__(self, keywords=None, percentile=0.95, use_mean=False):
        super().__init__(keywords)
        assert 0 < percentile <= 1, "Percentile must be in range (0,1]"
        self._percentile = percentile
        self._use_mean = use_mean

    def normalize(self, masked_kspace, image, **kwargs):
        magnitude_vals = cplx.abs(image).reshape(-1)
        k = int(round((1 - self._percentile) * magnitude_vals.numel()))
        scale = torch.min(torch.topk(magnitude_vals, k).values)

        outputs = {}
        outputs["masked_kspace"] = masked_kspace / scale
        outputs["image"] = image / scale
        if "target" in self._keywords:
            outputs["target"] = kwargs["target"] / scale

        mean = torch.tensor([0.0], dtype=torch.float32)
        std = scale.unsqueeze(-1)
        outputs.update({"mean": mean, "std": std})

        # Add other keys that were not computed.
        outputs.update({k: v for k, v in kwargs.items() if k not in outputs})
        return outputs


class Subsampler(object):
    def __init__(self, mask_func):
        self.mask_func = mask_func
        self.zip2_padding = None

    def _get_mask_shape(self, data_shape, mode: str):
        """Returns the shape of the mask based on the data shape.

        Args:
            data_shape (tuple[int]): The data shape.
            mode: Either ``"2D"`` or ``"3D"``
        """
        if mode == "2D":
            extra_dims = len(data_shape) - 3
            mask_shape = (1,) + data_shape[1:3] + (1,) * extra_dims
        elif mode == "3D":
            extra_dims = len(data_shape) - 4
            mask_shape = (1,) + data_shape[1:4] + (1,) * extra_dims
        else:
            raise ValueError("Only 2D and 3D undersampling masks are supported.")
        return mask_shape

    def __call__(self, data, mode: str = "2D", seed: int = None, acceleration: int = None):
        assert mode in ["2D", "3D"]
        data_shape = tuple(data.shape)
        if self.zip2_padding:
            data_shape = (
                data_shape[:1]
                + tuple(
                    s - 2 * p if p is not None else s
                    for s, p in zip(data_shape[1:], self.zip2_padding)
                )
                + data_shape[len(self.zip2_padding) + 1 :]
            )

        mask_shape = self._get_mask_shape(data_shape, mode)
        mask = self.mask_func(mask_shape, seed, acceleration)
        if self.zip2_padding:
            padded_mask_shape = self._get_mask_shape(tuple(data.shape), mode)
            padded_mask_shape = padded_mask_shape[1 : len(self.zip2_padding) + 1]
            mask = T.zero_pad(mask, padded_mask_shape)
        return torch.where(mask == 0, torch.tensor([0], dtype=data.dtype), data), mask


class DataTransform:
    """
    Data Transformer for training unrolled reconstruction models.

    For scans that
    """

    def __init__(
        self,
        cfg,
        mask_func,
        is_test: bool = False,
        add_noise: bool = False,
        add_motion: bool = False,
    ):
        """
        Args:
            mask_func (utils.subsample.MaskFunc): A function that can create a
                mask of appropriate shape.
            is_test (bool): If `True`, this class behaves with test-time
                functionality. In particular, it computes a pseudo random number
                generator seed from the filename. This ensures that the same
                mask is used for all the slices of a given volume every time.
        """
        from meddlr.transforms.builtin.mri import MRIReconAugmentor

        self._cfg = cfg
        self.mask_func = mask_func
        self._is_test = is_test

        # Build subsampler.
        # mask_func = build_mask_func(cfg)
        self._subsampler = Subsampler(self.mask_func)
        self.add_noise = add_noise
        self.add_motion = add_motion
        seed = cfg.SEED if cfg.SEED > -1 else None
        self.rng = np.random.RandomState(seed)
        if is_test:
            # When we test we dont want to initialize with certain parameters (e.g. scheduler).
            self.noiser = NoiseModel(cfg.MODEL.CONSISTENCY.AUG.NOISE.STD_DEV, seed=seed)
            self.motion_simulator = MotionModel(cfg.MODEL.CONSISTENCY.AUG.MOTION.RANGE, seed=seed)
        else:
            self.noiser = NoiseModel.from_cfg(cfg, seed=seed)
            self.motion_simulator = MotionModel.from_cfg(cfg, seed=seed)
        self.p_noise = cfg.AUG_TRAIN.NOISE_P
        self.p_motion = cfg.AUG_TRAIN.MOTION_P
        self._normalizer = build_normalizer(cfg)

        # Build augmentation pipeline.
        self.augmentor = None
        if not is_test and cfg.AUG_TRAIN.MRI_RECON.TRANSFORMS:
            self.augmentor = MRIReconAugmentor.from_cfg(cfg, aug_kind="aug_train", seed=seed)

    def _call_augmentor(
        self, kspace, maps, target, fname, slice_id, is_fixed, acceleration: int = None
    ):
        assert not self._is_test, "Augmentor is not supported with testing yet"

        # Convert everything from numpy arrays to tensors
        kspace = cplx.to_tensor(kspace).unsqueeze(0)
        maps = cplx.to_tensor(maps).unsqueeze(0)
        target_init = cplx.to_tensor(target).unsqueeze(0)
        target = (
            torch.complex(target_init, torch.zeros_like(target_init)).unsqueeze(-1)
            if not torch.is_complex(target_init)
            else target_init
        )  # handle rss vs. sensitivity-integrated
        norm = torch.sqrt(torch.mean(cplx.abs(target) ** 2))

        seed = sum(tuple(map(ord, fname))) if self._is_test or is_fixed else None  # noqa
        mask_gen = partial(
            self._subsampler.__call__, mode="2D", seed=seed, acceleration=acceleration
        )

        out, _, _ = self.augmentor(
            kspace,
            maps=maps,
            target=target,
            normalizer=self._normalizer,
            mask_gen=mask_gen,
            skip_tfm=is_fixed,  # Skip augmentations for unsupervised scans.
        )
        masked_kspace = out["kspace"]
        maps = out["maps"]
        target = out["target"]
        mean = out["mean"]
        std = out["std"]

        # Get rid of batch dimension...
        masked_kspace = masked_kspace.squeeze(0)
        maps = maps.squeeze(0)
        target = target.squeeze(0)

        return masked_kspace, maps, target, mean, std, norm

    def __call__(self, kspace, maps, target, fname, slice_id, is_fixed, acceleration: int = None):
        """
        Args:
            kspace (numpy.array): Input k-space of shape
                (num_coils, rows, cols, 2) for multi-coil
                data or (rows, cols, 2) for single coil data.
            target (numpy.array): Target image
            attrs (dict): Acquisition related information stored in the HDF5
                object.
            fname (str): File name
            slice (int): Serial number of the slice.
            is_fixed (bool, optional): If `True`, transform the example
                to have a fixed mask and acceleration factor.
            acceleration (int): Acceleration factor. Must be provided if
                `is_undersampled=True`.
        Returns:
            (tuple): tuple containing:
                image (torch.Tensor): Zero-filled input image.
                target (torch.Tensor): Target image converted to a torch Tensor.
                mean (float): Mean value used for normalization.
                std (float): Standard deviation value used for normalization.
                norm (float): L2 norm of the entire volume.
        """
        if is_fixed and not acceleration:
            raise ValueError("Accelerations must be specified for undersampled scans")

        # If augmentor is defined, use it to do computation.
        if self.augmentor is not None:
            return self._call_augmentor(
                kspace, maps, target, fname, slice_id, is_fixed, acceleration
            )

        # Convert everything from numpy arrays to tensors
        kspace = cplx.to_tensor(kspace).unsqueeze(0)
        maps = cplx.to_tensor(maps).unsqueeze(0)
        target_init = cplx.to_tensor(target).unsqueeze(0)
        target = (
            torch.complex(target_init, torch.zeros_like(target_init)).unsqueeze(-1)
            if not torch.is_complex(target_init)
            else target_init
        )  # handle rss vs. sensitivity-integrated
        norm = torch.sqrt(torch.mean(cplx.abs(target) ** 2))

        # TODO: Add other transforms here.

        # Apply mask in k-space
        seed = sum(tuple(map(ord, fname))) if self._is_test or is_fixed else None  # noqa
        masked_kspace, mask = self._subsampler(
            kspace, mode="2D", seed=seed, acceleration=acceleration
        )

        # Zero-filled Sense Recon.
        if torch.is_complex(target_init):
            A = SenseModel(maps, weights=mask)
            image = A(masked_kspace, adjoint=True)
        # Zero-filled RSS Recon.
        else:
            image = T.ifft2(masked_kspace)
            image_rss = torch.sqrt(torch.sum(cplx.abs(image) ** 2, axis=-1))
            image = torch.complex(image_rss, torch.zeros_like(image_rss)).unsqueeze(-1)

        # Normalize
        normalized = self._normalizer.normalize(
            **{"masked_kspace": masked_kspace, "image": image, "target": target, "mask": mask}
        )
        masked_kspace = normalized["masked_kspace"]
        target = normalized["target"]
        mean = normalized["mean"]
        std = normalized["std"]

        add_noise = self.add_noise and (
            self._is_test or (not is_fixed and self.rng.uniform() < self.p_noise)
        )
        add_motion = self.add_motion and (
            self._is_test or (not is_fixed and self.rng.uniform() < self.p_motion)
        )
        if add_noise:
            # Seed should be different for each slice of a scan.
            noise_seed = seed + slice_id if seed is not None else None
            masked_kspace = self.noiser(masked_kspace, mask=mask, seed=noise_seed)
        if add_motion:
            # Motion seed should not be different for each slice for now.
            # TODO: Change this for 2D acquisitions.
            masked_kspace = self.motion_simulator(masked_kspace, seed=seed)
        # Get rid of batch dimension...
        masked_kspace = masked_kspace.squeeze(0)
        maps = maps.squeeze(0)
        target = target.squeeze(0)

        return masked_kspace, maps, target, mean, std, norm



from numbers import Number
from typing import Dict, Sequence, Tuple, Union

import torch

from meddlr.transforms.base.spatial import (
    AffineTransform,
    FlipTransform,
    Rot90Transform,
    TranslationTransform,
)
from meddlr.transforms.build import TRANSFORM_REGISTRY
from meddlr.transforms.param_kind import ParamKind
from meddlr.transforms.transform import NoOpTransform
from meddlr.transforms.transform_gen import TransformGen

__all__ = ["RandomAffine", "RandomFlip", "RandomRot90", "RandomTranslation"]

SPATIAL_RANGE_OR_VAL = Union[float, Sequence[float], Sequence[Tuple[float, float]]]


@TRANSFORM_REGISTRY.register()
class RandomAffine(TransformGen):
    _base_transform = AffineTransform
    _param_names = ("angle", "translate", "scale", "shear")

    def __init__(
        self,
        p: Union[float, Dict[str, float]] = 0.0,
        angle: Union[float, Tuple[float, float]] = None,
        translate: SPATIAL_RANGE_OR_VAL = None,
        scale: Union[float, Tuple[float, float]] = None,
        shear: SPATIAL_RANGE_OR_VAL = None,
        pad_like=None,
    ):
        if isinstance(p, Number):
            p = {n: p for n in self._param_names}
        else:
            assert isinstance(p, dict)
            unknown_keys = set(p.keys()) - set(self._param_names)
            if len(unknown_keys):
                raise ValueError(f"Unknown keys for `p`: {unknown_keys}")
            p = p.copy()
            p.update({k: 0.0 for k in self._param_names if k not in p})
        params = locals()
        params = {k: params[k] for k in list(self._param_names)}
        self.pad_like = pad_like
        super().__init__(
            params=params,
            p=p,
            param_kinds={"translate": ParamKind.MULTI_ARG, "shear": ParamKind.MULTI_ARG},
        )

    def _get_params(self, shape):
        ndim = len(shape)
        params = self._get_param_values()

        p = params["p"]
        param_angle = params["angle"]
        param_translate = params["translate"]
        param_scale = params["scale"]
        param_shear = params["shear"]

        if isinstance(param_angle, Number):
            param_angle = (-param_angle, param_angle)
        if isinstance(param_translate, Number):
            param_translate = ((-param_translate, param_translate),)
        if isinstance(param_scale, Number):
            param_scale = tuple(sorted([1.0 - param_scale, 1.0 + param_scale]))
        if isinstance(param_shear, Number):
            param_shear = ((-param_shear, param_shear),)

        param_translate = self._format_param(param_translate, ParamKind.MULTI_ARG, ndim)
        param_shear = self._format_param(param_shear, ParamKind.MULTI_ARG, ndim)

        angle, translate, scale, shear = None, None, None, None

        if param_angle is not None and self._rand() < p["angle"]:
            angle = self._rand_range(*param_angle)
        if param_translate is not None and self._rand() < p["translate"]:
            translate = [int(self._rand_range(*x) * s) for x, s in zip(param_translate, shape)]
        if param_scale is not None and self._rand() < p["scale"]:
            scale = self._rand_range(*param_scale)
        if param_shear is not None and self._rand() < p["shear"]:
            shear = [self._rand_range(*x) for x in param_shear]

        return angle, translate, scale, shear

    def get_transform(self, image):
        # Affine only supports 2D spatial transforms
        spatial_shape = image.shape[-2:]

        out = self._get_params(spatial_shape)
        if all(x is None for x in out):
            return NoOpTransform()

        angle, translate, scale, shear = out
        return AffineTransform(angle=angle, translate=translate, scale=scale, shear=shear)


@TRANSFORM_REGISTRY.register()
class RandomTranslation(TransformGen):
    _base_transform = TranslationTransform

    def __init__(
        self,
        p: Union[float, Dict[str, float]] = 0.0,
        translate: SPATIAL_RANGE_OR_VAL = None,
        pad_mode=None,
        pad_value=0,
        ndim=2,
    ):
        params = {"translate": translate}
        self.pad_mode = pad_mode
        self.pad_value = pad_value
        self.ndim = ndim
        super().__init__(params=params, p=p, param_kinds={"translate": ParamKind.MULTI_ARG})

    def get_transform(self, image):
        shape = image.shape[-self.ndim :]
        ndim = len(shape)

        params = self._get_param_values(use_schedulers=True)
        p = params["p"]
        param_translate = params["translate"]
        translate = self._format_param(param_translate, ParamKind.MULTI_ARG, ndim)

        if self._rand() >= p:
            return NoOpTransform()
        translate = [int(self._rand_range(*x) * s) for x, s in zip(translate, shape)]
        return TranslationTransform(translate, pad_mode=self.pad_mode, pad_value=self.pad_value)


@TRANSFORM_REGISTRY.register()
class RandomFlip(TransformGen):
    _base_transform = FlipTransform

    def __init__(self, dims=None, ndim=None, p: Union[float, Dict[int, float]] = 0.0) -> None:
        if dims is None and ndim is None:
            raise ValueError("Either `dims` or `ndim` must be specified")
        if all(x is not None for x in (dims, ndim)):
            raise ValueError("Only one of `dims` or `ndim` can be specified.")
        if isinstance(dims, int):
            dims = (dims,)
        self.dims = dims
        self.ndim = ndim
        super().__init__(p=p)

    def get_transform(self, input):
        params = self._get_param_values(use_schedulers=True)
        p = params["p"]
        if self.dims is not None:
            dims = tuple(d for d in self.dims if self._rand() < p)
        else:
            if isinstance(p, Dict):
                dims = tuple(k for k, v in p.items() if self._rand() < v)
            else:
                dims = tuple(d for d in range(-self.ndim, 0) if self._rand() < p)

        return FlipTransform(dims) if dims else NoOpTransform()


@TRANSFORM_REGISTRY.register()
class RandomRot90(TransformGen):
    _base_transform = Rot90Transform

    def __init__(self, ks=None, p=0.0) -> None:
        self.ks = ks if ks is not None else list(range(1, 4))
        super().__init__(p=p)

    def get_transform(self, input):
        params = self._get_param_values(use_schedulers=True)
        if self._rand() >= params["p"]:
            return NoOpTransform()
        k = self.ks[torch.randperm(len(self.ks))[0].item()]
        return Rot90Transform(k=k, dims=(-1, -2))


def _duplicate_ndim(param, ndim):
    if param is None:
        return None

    if isinstance(param, Sequence) and isinstance(param[0], Sequence):
        return [[x if len(x) > 1 else (-x[0], x[0]) for x in y] for y in param]

    if isinstance(param, Sequence):
        param = (-param[0], param[0]) if len(param) == 1 else param
    else:
        param = (-param, param)
    return [param] * ndim



"""This file registers pre-defined datasets at hard-coded paths.

In the future, will also include metadata.

We hard-code some paths to the dataset that's assumed to
exist at the user's preferred dataset location (see utils/cluster.py).
"""

import os

from .register_mrco import register_mrco_scans

# ==== Predefined datasets and splits for MRCO formatted datasets ==========

_PREDEFINED_SPLITS_MRCO = {}
_PREDEFINED_SPLITS_MRCO["mridata_knee_2019"] = {
    "mridata_knee_2019_train": ("mridata_knee_2019/train", "ann://mridata_knee_2019/train.json"),
    "mridata_knee_2019_val": ("mridata_knee_2019/val", "ann://mridata_knee_2019/val.json"),
    "mridata_knee_2019_test": ("mridata_knee_2019/test", "ann://mridata_knee_2019/test.json"),
}


_PREDEFINED_SPLITS_MRCO["fastMRI_knee_multicoil"] = {
    # fastMRI knee multicoil toy dataset - do not use.
    "fastMRI_knee_multicoil_dev_train": (
        None,
        "fastmri/knee_multicoil/annotations/vtoy-dev/train.json",
    ),
    "fastMRI_knee_multicoil_dev_val": (
        None,
        "fastmri/knee_multicoil/annotations/vtoy-dev/val.json",
    ),
    "fastMRI_knee_multicoil_dev_test": (
        None,
        "fastmri/knee_multicoil/annotations/vtoy-dev/test.json",
    ),
    # fastMRI knee multicoil base v0.0.1 dataset.
    "fastMRI_knee_multicoil_v0.0.1_train": (
        "fastmri/knee_multicoil/train",
        "ann://fastmri/knee_multicoil/v0.0.1-dev/train.json",
    ),
    "fastMRI_knee_multicoil_v0.0.1_val": (
        "fastmri/knee_multicoil/train",
        "ann://fastmri/knee_multicoil/v0.0.1-dev/val.json",
    ),
    "fastMRI_knee_multicoil_v0.0.1_test": (
        "fastmri/knee_multicoil/val",
        "ann://fastmri/knee_multicoil/v0.0.1-dev/test.json",
    ),
    # fastMRI knee multicoil mini v0.0.1 dataset.
    "fastMRI_knee_multicoil_mini_v0.0.1_train": (
        "fastmri/knee_multicoil/val",
        "ann://fastmri/knee_multicoil/mini-v0.0.1/train.json",
    ),
    "fastMRI_knee_multicoil_mini_v0.0.1_val": (
        "fastmri/knee_multicoil/val",
        "ann://fastmri/knee_multicoil/mini-v0.0.1/val.json",
    ),
    "fastMRI_knee_multicoil_mini_v0.0.1_test": (
        "fastmri/knee_multicoil/val",
        "ann://fastmri/knee_multicoil/mini-v0.0.1/test.json",
    ),
    # fastMRI knee multicoil ultra mini v0.0.1 dataset.
    "fastMRI_knee_multicoil_ultra_mini_v0.0.1_train": (
        "fastmri/knee_multicoil/val",
        "ann://fastmri/knee_multicoil/ultra-mini-v0.0.1/train.json",
    ),
    "fastMRI_knee_multicoil_ultra_mini_v0.0.1_val": (
        "fastmri/knee_multicoil/val",
        "ann://fastmri/knee_multicoil/ultra-mini-v0.0.1/val.json",
    ),
    "fastMRI_knee_multicoil_ultra_mini_v0.0.1_test": (
        "fastmri/knee_multicoil/val",
        "ann://fastmri/knee_multicoil/ultra-mini-v0.0.1/test.json",
    ),
    # fastMRI knee multicoil - fat-suppressed
    "fastMRI_knee_multicoil_v0.0.1_fs_train": (
        "fastmri/knee_multicoil/train",
        "ann://fastmri/knee_multicoil/v0.0.1-dev-fs/train.json",
    ),
    "fastMRI_knee_multicoil_v0.0.1_fs_val": (
        "fastmri/knee_multicoil/train",
        "ann://fastmri/knee_multicoil/v0.0.1-dev-fs/val.json",
    ),
    "fastMRI_knee_multicoil_v0.0.1_fs_test": (
        "fastmri/knee_multicoil/val",
        "ann://fastmri/knee_multicoil/v0.0.1-dev-fs/test.json",
    ),
    # fastMRI knee multicoil - 3T fat-suppressed
    "fastMRI_knee_multicoil_v0.0.1_fs_3t_train": (
        "fastmri/knee_multicoil/train",
        "ann://fastmri/knee_multicoil/v0.0.1-dev-fs-3t/train.json",
    ),
    "fastMRI_knee_multicoil_v0.0.1_fs_3t_val": (
        "fastmri/knee_multicoil/train",
        "ann://fastmri/knee_multicoil/v0.0.1-dev-fs-3t/val.json",
    ),
    "fastMRI_knee_multicoil_v0.0.1_fs_3t_test": (
        "fastmri/knee_multicoil/val",
        "ann://fastmri/knee_multicoil/v0.0.1-dev-fs-3t/test.json",
    ),
}

_PREDEFINED_SPLITS_MRCO["fastMRI_brain_multicoil"] = {
    "fastMRI_brain_multicoil_dev_train": (
        "fastmri/brain_multicoil/train",
        "ann://fastmri/brain_multicoil/vtoy-dev/train.json",
    ),
    "fastMRI_brain_multicoil_dev_val": (
        "fastmri/brain_multicoil/train",
        "ann://fastmri/brain_multicoil/vtoy-dev/val.json",
    ),
    "fastMRI_brain_multicoil_dev_test": (
        "fastmri/brain_multicoil/val",
        "ann://fastmri/brain_multicoil/vtoy-dev/test.json",
    ),
    # Mini split
    "fastMRI_brain_multicoil_mini_v0.0.1_train": (
        "fastmri/brain_multicoil/val",
        "ann://fastmri/brain_multicoil/mini-v0.0.1/train.json",
    ),
    "fastMRI_brain_multicoil_mini_v0.0.1_val": (
        "fastmri/brain_multicoil/val",
        "ann://fastmri/brain_multicoil/mini-v0.0.1/val.json",
    ),
    "fastMRI_brain_multicoil_mini_v0.0.1_test": (
        "fastmri/brain_multicoil/val",
        "ann://fastmri/brain_multicoil/mini-v0.0.1/test.json",
    ),
}


_METADATA_FILES = {
    "mridata_knee_2019": "ann://mridata_knee_2019/metadata.csv",
    "fastMRI_knee_multicoil": "ann://fastmri/knee_multicoil/fastmri_knee_multicoil_metadata.csv",
    "fastMRI_brain_multicoil": "ann://fastmri/brain_multicoil/fastmri_brain_multicoil_metadata.csv",
}


def register_all_mrco(root="data://"):
    for dataset_name, splits_per_dataset in _PREDEFINED_SPLITS_MRCO.items():
        metadata_file = _METADATA_FILES.get(dataset_name, None)
        for key, data in splits_per_dataset.items():
            # Assume pre-defined datasets live in `./datasets`.
            image_root, json_file = data

            json_file = os.path.join(root, json_file) if "://" not in json_file else json_file
            image_root = os.path.join(root, image_root) if image_root is not None else None
            register_mrco_scans(
                key,
                {"metadata_file": metadata_file},  # TODO: add metadata
                json_file,  # noqa
                image_root,
            )


# Register them all under "data://"
register_all_mrco()



from typing import Tuple, Union

import torch.nn as nn
import torch.nn.functional as F

from meddlr.modeling.layers.build import CUSTOM_LAYERS_REGISTRY

__all__ = ["ConvWS2d", "ConvWS3d"]


@CUSTOM_LAYERS_REGISTRY.register()
class ConvWS2d(nn.Conv2d):
    """Conv2d with Weight Standardization.

    Adapted from
    https://github.com/joe-siyuan-qiao/pytorch-classification/blob/e6355f829e85ac05a71b8889f4fff77b9ab95d0b/models/layers.py

    References:
        S. Qiao, et al. Micro-Batch Training with Batch-Channel Normalization and Weight
        Standardization, ArXiv 2020. https://arxiv.org/abs/1903.10520
    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: Union[int, Tuple[int, int]],
        stride=1,
        padding=0,
        dilation=1,
        groups=1,
        bias=True,
        eps=1e-5,
    ):
        super().__init__(
            in_channels, out_channels, kernel_size, stride, padding, dilation, groups, bias
        )
        self.eps = eps

    def forward(self, x):
        weight = self.weight
        weight_mean = weight.view(weight.size(0), -1).mean(dim=1).view(-1, 1, 1, 1)
        weight = weight - weight_mean
        std = weight.view(weight.size(0), -1).std(dim=1).view(-1, 1, 1, 1) + self.eps
        weight = weight / std.expand_as(weight)
        return F.conv2d(x, weight, self.bias, self.stride, self.padding, self.dilation, self.groups)


@CUSTOM_LAYERS_REGISTRY.register()
class ConvWS3d(nn.Conv3d):
    """Conv3d with Weight Standardization.

    Adapted from:
    https://github.com/joe-siyuan-qiao/pytorch-classification/blob/e6355f829e85ac05a71b8889f4fff77b9ab95d0b/models/layers.py

    References:
        S. Qiao, et al. Micro-Batch Training with Batch-Channel Normalization and Weight
        Standardization, ArXiv 2020. https://arxiv.org/abs/1903.10520
    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: Union[int, Tuple[int, int]],
        stride=1,
        padding=0,
        dilation=1,
        groups=1,
        bias=True,
        eps=1e-5,
    ):
        super().__init__(
            in_channels, out_channels, kernel_size, stride, padding, dilation, groups, bias
        )
        self.eps = eps

    def forward(self, x):
        weight = self.weight
        weight_mean = weight.view(weight.size(0), -1).mean(dim=1).view(-1, 1, 1, 1, 1)
        weight = weight - weight_mean
        std = weight.view(weight.size(0), -1).std(dim=1).view(-1, 1, 1, 1, 1) + self.eps
        weight = weight / std.expand_as(weight)
        return F.conv3d(x, weight, self.bias, self.stride, self.padding, self.dilation, self.groups)



import contextlib
from typing import Dict, Sequence, Tuple, Union

import h5py
import numpy as np
import torch

from meddlr.utils import profiler

__all__ = ["HDF5Manager", "structure_patches", "collect_mask"]


class HDF5Manager:
    """Manager for opening and caching HDF5 files."""

    def __init__(self, files: Sequence[str] = None, cache=True):
        self.files: Dict[str, h5py.File] = {}
        self.cache = cache
        if files and cache:
            for fp in files:
                self.open(fp)

    def open(self, filepath):
        """Open and cache file in read mode.
        Args:
            filepath (str): Path to HDF5 file
        """
        if filepath in self.files:
            return
        self.files[filepath] = h5py.File(filepath, "r")

    def close(self, filepath, drop=True):
        """Open and cache file in read mode.
        Args:
            filepath (str): Path to HDF5 file
            drop (bool, optional): If `True`, drop filepath from list of files.
                If `False`, `self.files` will keep a closed copy of the file.
        """
        self.files[filepath].close()
        if drop:
            self.files.pop(filepath)

    @profiler.time_profile()
    def get(self, filepath: str, key: str = None, patch: Union[str, Sequence[slice]] = None):
        """Get data from h5 file.
        Args:
            filepath (str): Filepath to fetch data from.
            key (str, optional): HDF5 key. If `None`, returns open file.
            patch ()
        """

        if patch is not None and key is None:
            raise ValueError("`key` must be specified to use `patch`.")

        if filepath in self.files:
            file = self.files[filepath]
            return self._load_data(file, key=key, patch=patch)
        else:
            assert key
            with h5py.File(filepath, "r") as file:
                data = self._load_data(file, key=key, patch=patch)
            return data

    @profiler.time_profile()
    @contextlib.contextmanager
    def yield_file(self, filepath):
        is_cached = filepath in self.files
        file = self.files[filepath] if is_cached else h5py.File(filepath, "r")
        if self.cache and not is_cached:
            self.files[filepath] = file

        try:
            yield file
        finally:
            if not is_cached and not self.cache:
                file.close()

    def _load_data(self, file, key=None, patch=None):
        if not key:
            return file

        val = file[key]
        if patch is None:
            return val

        if isinstance(patch, str):
            assert patch in ["all", "load"]
            patch = ()
        elif not isinstance(patch, tuple):
            patch = tuple(patch)
        return val[patch]

    def __del__(self):
        # Copy keys to allow modifying self.files dict.
        for filepath in list(self.files.keys()):
            self.close(filepath)


def structure_patches(
    patches: Sequence[torch.Tensor],
    coords: Union[Sequence[int], Sequence[Tuple[int]]] = None,
    dims=None,
):
    if coords is None:
        coords = list(patches.keys())
        patches = list(patches.values())

    if len(patches) != len(coords):
        raise ValueError("Got unknown length for coordinates")

    if isinstance(coords[0], int):
        coords = [(x,) for x in coords]

    if len(coords) != len(set(coords)):
        raise ValueError(
            "All coordinates must be unique. " "Overlapping patches are not currently supported."
        )

    if dims is not None:
        if isinstance(dims, int):
            dims = (dims,)
        if len(dims) != len(coords[0]):
            raise ValueError(
                f"Specified {len(dims)} dimensions, but have {len(coords[0])}-d coordinates"
            )
        if len(dims) != len(set(dims)):
            raise ValueError(f"Duplicate dimensions specified - {dims}")

    base_shape = patches[0].shape
    if any(x.shape != base_shape for x in patches):
        raise ValueError("All patches must have the same shape")

    stack_shape = []
    coords_arr = np.asarray(coords)
    assert coords_arr.ndim == 2
    stack_shape = tuple(np.max(coords_arr[:, c]).item() + 1 for c in range(coords_arr.shape[1]))

    struct = np.empty(stack_shape, dtype=np.object)
    for coord, patch in zip(coords, patches):
        struct[coord] = patch
    if any(x is None for x in struct.flatten()):
        raise ValueError("Did not get dense labels.")
    struct = struct.tolist()

    out = _recursive_stack(struct)

    if dims is not None:
        # Reorder dimensions of the structured tensor.
        dims = [out.ndim + d if d < 0 else d for d in dims]
        original_dims_new_order = sorted(set(range(out.ndim)) - set(dims))
        order_dict = {v: k for k, v in zip(range(out.ndim), dims + original_dims_new_order)}
        out = out.permute(tuple(order_dict[k] for k in sorted(order_dict.keys())))

    return out


def collect_mask(
    mask: np.ndarray,
    index: Sequence[Union[int, Sequence[int], int]],
    out_channel_first: bool = True,
):
    """Collect masks by index.

    Collated indices will be summed. For example, `index=(1,(3,4))` will return
    `np.stack(mask[...,1], mask[...,3]+mask[...,4])`.

    TODO: Add support for adding background.

    Args:
        mask (ndarray): A (...)xC array.
        index (Sequence[int]): The index/indices to select in mask.
            If sub-indices are collated, they will be summed.
        out_channel_first (bool, optional): Reorders dimensions of output mask to Cx(...)
    """
    if isinstance(index, int):
        index = (index,)

    if not any(isinstance(idx, Sequence) for idx in index):
        mask = mask[..., index]
    else:
        o_seg = []
        for idx in index:
            c_seg = mask[..., idx]
            if isinstance(idx, Sequence):
                c_seg = np.sum(c_seg, axis=-1)
            o_seg.append(c_seg)
        mask = np.stack(o_seg, axis=-1)

    if out_channel_first:
        last_idx = len(mask.shape) - 1
        mask = np.transpose(mask, (last_idx,) + tuple(range(0, last_idx)))

    return mask


def _recursive_stack(tensors) -> torch.Tensor:
    if all(isinstance(x, torch.Tensor) for x in tensors):
        return torch.stack(tensors, dim=0)
    if all(isinstance(x, list) for x in tensors):
        tensors = [_recursive_stack(x) for x in tensors]
        return torch.stack(tensors, dim=0)
    else:
        raise ValueError("Unknown stack")



"""Implementation of 2D base layers.
"""

from typing import Tuple, Union

from torch import nn


def _get_same_padding(kernel_size: Union[int, Tuple[int, int]]):
    if isinstance(kernel_size, int):
        kernel_size = (kernel_size, kernel_size)
    else:
        assert len(kernel_size) == 2
    if not all(k % 2 == 1 for k in kernel_size):
        raise ValueError("Kernel size must be odd - got {}".format(kernel_size))
    padding = tuple(k // 2 for k in kernel_size)

    return padding


class ConvBlock(nn.Module):
    """
    A 2D Convolutional Block that consists of Norm -> ReLU -> Dropout -> Conv

    Based on implementation described by:
        K He, et al. "Identity Mappings in Deep Residual Networks" arXiv:1603.05027
    """

    def __init__(
        self,
        in_chans: int,
        out_chans: int,
        kernel_size: Union[int, Tuple[int, int]],
        drop_prob: float,
        act_type: str = "relu",
        norm_type: str = "none",
        norm_affine: bool = False,
        order: Tuple[str, str, str, str] = ("norm", "act", "drop", "conv"),
    ):
        """
        Args:
            in_chans (int): Number of channels in the input.
            out_chans (int): Number of channels in the output.
            drop_prob (float): Dropout probability.
        """
        super().__init__()

        self.in_chans = in_chans
        self.out_chans = out_chans
        self.drop_prob = drop_prob

        if isinstance(kernel_size, int):
            kernel_size = (kernel_size, kernel_size)
        else:
            assert len(kernel_size) == 2
        if not all(k % 2 == 1 for k in kernel_size):
            raise ValueError("Kernel size must be odd - got {}".format(kernel_size))

        padding = tuple(k // 2 for k in kernel_size)

        # Define choices for each layer in ConvBlock
        conv_after_norm = "norm" in order and order.index("conv") > order.index("norm")
        norm_channels = in_chans if conv_after_norm else out_chans
        normalizations = nn.ModuleDict(
            [
                ["none", nn.Identity()],
                ["instance", nn.InstanceNorm2d(norm_channels, affine=norm_affine)],
                ["batch", nn.BatchNorm2d(norm_channels, affine=norm_affine)],
            ]
        )
        activations = nn.ModuleDict([["relu", nn.ReLU()], ["leaky_relu", nn.LeakyReLU()]])
        dropout = nn.Dropout2d(p=drop_prob)
        convolution = nn.Conv2d(in_chans, out_chans, kernel_size=kernel_size, padding=padding)

        layer_dict = {
            "conv": convolution,
            "drop": dropout,
            "act": activations[act_type],
            "norm": normalizations[norm_type] if norm_type in normalizations else nn.Identity(),
        }
        layers = [layer_dict[lyr] for lyr in order]

        # Define forward pass
        self.layers = nn.Sequential(*layers)

    def forward(self, input):
        """
        Args:
            input (torch.Tensor): Input tensor of shape ``(B,C_{in},D,H,W)``.

        Returns:
            (torch.Tensor): Output tensor of shape ``(B,C_{in},D,H,W)``.
        """
        return self.layers(input)


class ResBlock(nn.Module):
    """
    A ResNet block that consists of two convolutional layers
    followed by a residual connection.
    """

    def __init__(
        self,
        in_chans,
        out_chans,
        kernel_size,
        drop_prob,
        act_type: str = "relu",
        norm_type: str = "none",
        norm_affine: bool = False,
        order: Tuple[str, str, str, str] = ("norm", "act", "drop", "conv"),
    ):
        """
        Args:
            in_chans (int): Number of channels in the input.
            out_chans (int): Number of channels in the output.
            drop_prob (float): Dropout probability.
        """
        super().__init__()

        self.layers = nn.Sequential(
            ConvBlock(
                in_chans, out_chans, kernel_size, drop_prob, act_type, norm_type, norm_affine, order
            ),  # noqa
            ConvBlock(
                out_chans,
                out_chans,
                kernel_size,
                drop_prob,
                act_type,
                norm_type,
                norm_affine,
                order,
            ),  # noqa
        )

        if in_chans != out_chans:
            self.resample = nn.Conv2d(in_chans, out_chans, kernel_size=1)
        else:
            self.resample = nn.Identity()

    def forward(self, input):
        """
        Args:
            input (torch.Tensor): Input tensor of shape ``(B,C_{in},D,H,W)``.

        Returns:
            (torch.Tensor): Output tensor of shape ``(B,C_{in},D,H,W)``.
        """

        # To have a residual connection, number of inputs must be equal to outputs
        shortcut = self.resample(input)

        return self.layers(input) + shortcut


class ResNet(nn.Module):
    """
    Prototype for 3D ResNet architecture
    """

    def __init__(
        self,
        num_resblocks,
        in_chans,
        chans,
        kernel_size,
        drop_prob,
        circular_pad=False,
        act_type: str = "relu",
        norm_type: str = "none",
        norm_affine: bool = False,
        order: Tuple[str, str, str, str] = ("norm", "act", "drop", "conv"),
    ):
        """ """
        super().__init__()

        if circular_pad:
            raise NotImplementedError(
                "Circular padding is not available. "
                "It is retained in the init to be used in the future."
            )
        self.circular_pad = circular_pad
        self.pad_size = 2 * num_resblocks + 1

        resblock_params = {
            "act_type": act_type,
            "norm_type": norm_type,
            "norm_affine": norm_affine,
            "order": order,
            "kernel_size": kernel_size,
            "drop_prob": drop_prob,
        }
        # Declare ResBlock layers
        self.res_blocks = nn.ModuleList([ResBlock(in_chans, chans, **resblock_params)])
        for _ in range(num_resblocks - 1):
            self.res_blocks += [ResBlock(chans, chans, **resblock_params)]

        # Declare final conv layer (down-sample to original in_chans)
        padding = _get_same_padding(kernel_size)
        self.final_layer = nn.Conv2d(chans, in_chans, kernel_size=kernel_size, padding=padding)

    def forward(self, input):
        """
        Args:
            input (torch.Tensor): Input tensor of shape ``(B,C_{in},D,H,W)``.

        Returns:
            (torch.Tensor): Output tensor of shape ``(B,C_{in},D,H,W)``.
        """

        # orig_shape = input.shape
        # if self.circular_pad:
        #     input = nn.functional.pad(
        #         input, 2 * (self.pad_size,) + (0, 0), mode="circular"
        #     )

        # Perform forward pass through the network
        output = input
        for res_block in self.res_blocks:
            output = res_block(output)
        output = self.final_layer(output) + input

        # return center_crop(output, orig_shape)

        return output



import numpy as np
import torch
import torch.nn.functional as F

__all__ = [
    "one_hot_to_categorical",
    "categorical_to_one_hot",
    "logits_to_prob",
    "pred_to_categorical",
]


def one_hot_to_categorical(x, channel_dim: int = 1, background=False):
    """Converts one-hot encoded predictions to categorical predictions.

    Args:
        x (torch.Tensor | np.ndarray): One-hot encoded predictions.
        channel_dim (int, optional): Channel dimension.
            Defaults to ``1`` (i.e. ``(B,C,...)``).
        background (bool, optional): If ``True``, assumes index 0 in the
            channel dimension is the background.

    Returns:
        torch.Tensor | np.ndarray: Categorical array or tensor. If ``background=False``,
        the output will be 1-indexed such that ``0`` corresponds to the background.
    """
    is_ndarray = isinstance(x, np.ndarray)
    if is_ndarray:
        x = torch.as_tensor(x)

    if background is not None and background is not False:
        out = torch.argmax(x, channel_dim)
    else:
        out = torch.argmax(x.type(torch.long), dim=channel_dim) + 1
        out = torch.where(x.sum(channel_dim) == 0, torch.tensor([0], device=x.device), out)

    if is_ndarray:
        out = out.numpy()
    return out


def categorical_to_one_hot(x, channel_dim: int = 1, background=0, num_categories=None, dtype=None):
    """Converts categorical predictions to one-hot encoded predictions.

    Args:
        x (torch.Tensor | np.ndarray): Categorical array or tensor.
        channel_dim (int, optional): Channel dimension for output tensor.
        background (int | NoneType, optional): The numerical label of the
            background category. If ``None``, assumes that the background is
            a class that should be one-hot encoded.
        num_categories (int, optional): Number of categories (excluding background).
            Defaults to the ``max(x) + 1``.
        dtype (type, optional): Data type of the output.
            Defaults to boolean (``torch.bool`` or ``np.bool``).

    Returns:
        torch.Tensor | np.ndarray: One-hot encoded predictions.
    """
    is_ndarray = isinstance(x, np.ndarray)
    if is_ndarray:
        x = torch.from_numpy(x)

    if num_categories is None:
        num_categories = torch.max(x).type(torch.long).cpu().item()
    num_categories += 1

    shape = x.shape
    out_shape = (num_categories,) + shape

    if dtype is None:
        dtype = torch.bool
    default_value = True if dtype == torch.bool else 1
    if x.dtype != torch.long:
        x = x.type(torch.long)

    out = torch.zeros(out_shape, dtype=dtype, device=x.device)
    out.scatter_(0, x.reshape((1,) + x.shape), default_value)
    if background is not None:
        out = torch.cat([out[0:background], out[background + 1 :]], dim=0)
    if channel_dim != 0:
        if channel_dim < 0:
            channel_dim = out.ndim + channel_dim
        order = (channel_dim,) + tuple(d for d in range(out.ndim) if d != channel_dim)
        out = out.permute(tuple(np.argsort(order)))
        out = out.contiguous()

    if is_ndarray:
        out = out.numpy()
    return out


def logits_to_prob(logits, activation, channel_dim: int = 1):
    """Converts logits to probabilities.

    Args:
        logits (torch.Tensor | np.ndarray): The logits.
        activation (str): Activation to use. One of ``'sigmoid'``, ``'softmax'``.
        channel_dim (int, optional): The channel dimension.

    Returns:
        torch.Tensor | np.ndarray: The probabilities.
    """
    is_ndarray = isinstance(logits, np.ndarray)
    if is_ndarray:
        logits = torch.from_numpy(logits)

    if activation == "sigmoid":
        out = torch.sigmoid(logits)
    elif activation == "softmax":
        out = F.softmax(logits, dim=channel_dim)

    if is_ndarray:
        out = out.numpy()
    return out


def pred_to_categorical(pred_or_logits, activation, channel_dim: int = 1, threshold: float = 0.5):
    """Converts one-hot encoded predictions or logits to category.

    Args:
        pred_or_logits: One-hot encoded predictions or logits. Shape BxCx...
        activation (str): Activation to use.
            Either ``'sigmoid'`` or ``'softmax'`` if ``pred_or_logits`` are logits.
            If `None` or '', assumes that `pred` does not need to be passed through
            activation function. If 'softmax', should include a background class.
        include_background (bool): If `True`, the first slice of class dimension (``C``)
            will not be dropped.
    """
    if activation not in [None, "", "sigmoid", "softmax"]:
        raise ValueError(f"activation '{activation}' not supported'")

    pred = pred_or_logits
    is_ndarray = isinstance(pred, np.ndarray)
    if is_ndarray:
        pred = torch.from_numpy(pred)

    if activation == "sigmoid":
        # TODO: Validate this case.
        out = one_hot_to_categorical(torch.sigmoid(pred) > threshold, channel_dim=channel_dim)
    elif activation == "softmax":
        out = torch.argmax(pred, dim=channel_dim)
    elif activation in (None, ""):
        # if not activation specified, assume it is one-hot encoded.
        out = one_hot_to_categorical(pred, channel_dim=channel_dim)
    else:
        raise ValueError(f"activation '{activation}' not supported'")

    if is_ndarray:
        out = out.numpy()
    return out



"""Detect clusters and machines.

DO NOT MOVE THIS FILE.
"""
import os
import re
import socket
from typing import Any, Dict, List, Sequence, Union

import yaml

from meddlr.utils import env

_PATH_MANAGER = env.get_path_manager()

__all__ = ["Cluster", "set_cluster"]


class Cluster:
    """Manages configurations for different nodes/clusters.

    This class is helpful for managing different cluster configurations
    (e.g. storage paths) across different nodes/clusters without the
    overhead of duplicating the codebase across multiple nodes.

    A cluster is defined as a combination of a set of nodes (i.e. machines)
    that share configuration properties, such as storage paths. This class helps
    manage the configuration properties of these sets of nodes together. For example,
    let's say machines with hostnames *nodeA* and *nodeB* share some data and results
    paths. We can define a new cluster *MyCluster* to manage these:

    >>> cluster = Cluster(
    ...     'MyCluster', patterns=['nodeA', 'nodeB'],
    ...     data_dir="/path/to/datasets", results_dir="/path/to/results"
    ... )

    To use the configurations of a particular cluster, set the working cluster:

    >>> cluster.use()

    Configs can be persisted by saving to a file. If the config has been saved,
    future sessions will attempt to auto-detect the set of saved configs:

    >>> cluster.save()  # save cluster configuration to file
    >>> cluster.delete()  # deletes cluster configuration from file

    To get the file where the configs are stored, run:

    >>> Cluster.config_file()  # get the config file for the cluster

    To identify the cluster config to use, we inspect the hostname of the current node.
    This can be problematic if two machines have the same hostname, though
    this has not been an issue as of yet.

    Note:
        DO NOT use the machine's public ip address to identify it. While this is
        definitely more robust, there are security issues associated with this.

    Note:
        This class is not thread safe. Saving/deleting configs should be done on
        the main thread.
    """

    def __init__(
        self,
        name: str = None,
        patterns: Union[str, Sequence[str]] = None,
        data_dir: str = None,
        results_dir: str = None,
        cache_dir: str = None,
        **cfg_kwargs,
    ):
        """
        Args:
            name (str): The name of the cluster. Name is case-sensitive.
            patterns (Sequence[str]): Regex pattern(s) for identifying nodes
                in the cluster. Cluster will be identified by
                ``any(re.match(p, socket.gethostname()) for p in patterns)``.
                If ``None``, defaults to current hostname.
            data_dir (str, optional): The data directory. Defaults to
                ``os.environ['MEDDLR_DATASETS_DIR']`` or ``"./datasets"``.
            results_dir (str, optional): The results directory. Defaults to
                `"os.environ['MEDDLR_RESULTS_DIR']"` or ``"./results"``.
            cfg_kwargs (optional): Any other configurations you would like to
                store for the cluster.
        """
        if name is None:
            name = socket.gethostname()
        self.name = name

        if patterns is None:
            patterns = socket.gethostname()
        if isinstance(patterns, str):
            patterns = [patterns]
        self.patterns = patterns

        self._data_dir = data_dir
        self._results_dir = results_dir
        self._cache_dir = cache_dir
        self._cfg_kwargs = cfg_kwargs

    @property
    def data_dir(self):
        path = self._data_dir
        path = os.environ.get("MEDDLR_DATASETS_DIR", path if path else "./datasets")
        return _PATH_MANAGER.get_local_path(path)

    @property
    def datasets_dir(self):
        """Alias for ``self.data_dir``."""
        return self.data_dir

    @property
    def results_dir(self):
        path = self._results_dir
        path = os.environ.get("MEDDLR_RESULTS_DIR", path if path else "./results")
        return _PATH_MANAGER.get_local_path(path)

    @property
    def cache_dir(self):
        path = self._cache_dir
        path = os.environ.get("MEDDLR_CACHE_DIR", path if path else "~/cache/meddlr")
        return _PATH_MANAGER.get_local_path(path)

    def set(self, **kwargs):
        """Set cluster configuration properties.

        Args:
            kwargs: Keyword arguments to set.

        Examples:
            >>> cluster.set(data_dir="/path/to/datasets", results_dir="/path/to/results")

        Note:
            Setting attributes for the cluster will not override environment variables.
            For example, if ``os.environ['MEDDLR_DATASETS_DIR']`` is set, setting ``data_dir``
            will have no effect.
        """
        for k, v in kwargs.items():
            private_key = f"_{k}"
            if hasattr(self, private_key):
                setattr(self, private_key, v)
            else:
                self._cfg_kwargs[k] = v

    def __getattr__(self, attr: str):
        attr_env = f"MEDDLR_{attr.upper()}"
        try:
            value = os.environ.get(attr_env, self._cfg_kwargs[attr])
        except KeyError:
            raise AttributeError(f"Attribute {attr} not specified for cluster {self.name}.")
        return value

    def save(self):
        """Save cluster config to yaml file."""
        filepath = self.config_file()
        os.makedirs(os.path.dirname(filepath), exist_ok=True)

        cluster_cfgs = {}
        if os.path.isfile(filepath):
            with open(filepath, "r") as f:
                cluster_cfgs = yaml.safe_load(f)

        data = {(k[1:] if k.startswith("_") else k): v for k, v in self.__dict__.items()}
        cluster_cfgs[self.name] = data

        with open(filepath, "w") as f:
            yaml.safe_dump(cluster_cfgs, f)

    def delete(self):
        """Deletes cluster config from yaml file."""
        filepath = self.config_file()
        if not os.path.isfile(filepath):
            return

        with open(filepath, "r") as f:
            cluster_cfgs: Dict[str, Any] = yaml.safe_load(f)

        if self.name not in cluster_cfgs:
            return

        cluster_cfgs.pop(self.name)
        with open(filepath, "w") as f:
            yaml.safe_dump(cluster_cfgs, f)

    def get_path(self, key):
        """Legacy method for fetching cluster-specific paths."""
        return getattr(self, key)

    @classmethod
    def all_clusters(cls) -> List["Cluster"]:
        return cls.from_config(name=None)

    @classmethod
    def cluster(cls):
        """Searches saved clusters by regex matching with hostname.

        Returns:
            Cluster: The current cluster.

        Note:
            The cluster must have been saved to a config file. Also, if
            there are multiple cluster matches, only the first (sorted alphabetically)
            will be returned.
        """
        try:
            clusters = cls.all_clusters()
        except FileNotFoundError:
            return _UNKNOWN
        hostname = socket.gethostname()
        for clus in clusters:
            if any(re.match(p, hostname) for p in clus.patterns):
                return clus
        return _UNKNOWN

    @classmethod
    def from_config(cls, name):
        """Loads cluster from config.

        Args:
            name (str | Sequence[str] | None): Cluster name(s) to load.
                If ``None``, all clusters will be loaded.

        Returns:
            Cluster: The Cluster object(s).
        """
        filepath = cls.config_file()
        if not os.path.isfile(filepath):
            raise FileNotFoundError(f"Config file not found: {filepath}")

        with open(filepath, "r") as f:
            cfg = yaml.safe_load(f)

        if name is None:
            return [cls(**cluster_cfg) for cluster_cfg in cfg.values()]
        elif isinstance(name, str):
            return cls(**cfg[name])
        else:
            return type(name)([cls(**cfg[n]) for n in name])

    @staticmethod
    def config_file():
        return os.path.join(env.settings_dir(), "clusters.yaml")

    @staticmethod
    def working_cluster() -> "Cluster":
        return _CLUSTER

    def use(self):
        """Sets ``self`` to be the working cluster of the project.

        The working cluster is the default cluster that is used to manage
        paths and other configuration variables.

        Examples:
            >>> cluster.use()

        Note:
            This function does not override environment variables.
            All environment variables will take priority over this clusters
        """
        set_cluster(self)

    def __repr__(self):
        return "Cluster({})".format(
            ", ".join("{}={}".format(k, v) for k, v in self.__dict__.items())
        )


def set_cluster(cluster: Union[str, Cluster] = None):
    """Sets the working cluster.
    Args:
        cluster (`str` or `Cluster`): The cluster name or cluster.
            If ``None``, will reset cluster to _UNKNOWN, meaning default
            data and results dirs will be used.
    """
    if cluster is None:
        cluster = _UNKNOWN
    elif isinstance(cluster, str):
        if cluster.lower() == _UNKNOWN.name.lower():
            cluster = _UNKNOWN
        else:
            cluster = Cluster.from_config(cluster)
    global _CLUSTER
    _CLUSTER = cluster


_UNKNOWN = Cluster("UNKNOWN", [])  # Unknown cluster
_CLUSTER = Cluster.cluster()  # Working cluster



from typing import Sequence, Union

import numpy as np
import torch

from meddlr.utils.events import get_event_storage


class MotionModel:
    """A model that corrupts kspace inputs with motion.

    Motion is a common artifact experienced during the MR imaging forward problem.
    When a patient moves, the recorded (expected) location of the kspace sample is
    different than the actual location where the kspace sample that was acquired.
    This module is responsible for simulating different motion artifacts.

    Attributes:
        motion_range (Tuple[float]): The range of motion difficulty.
            If a single value is provided on initialization, it is stored
            as a tuple with 1 element.
        warmup_method (str): The method that is being used for warmup.
        warmup_iters (int): Number of iterations to use for warmup.
        generator (torch.Generator): The generator that should be used for all
            random logic in this class.

    Note:
        This class is functionally deprecated and will not be maintained.
        Use :cls:`RandomMRIMotion` instead.

    Note:
        We do not store this as a module or else it would be saved to the model
        definition, which we dont want.

    Note:
        There is a known bug that the warmup method does not clamp the upper
        bound at the appropriate value. Thus the upper bound of the range keeps
        growing. We have elected not to correct for this to preserve
        reproducibility for older results. To use schedulers with corrected
        functionality, see :cls:`RandomMRIMotion` instead.
    """

    def __init__(self, motion_range: Union[float, Sequence[float]], scheduler=None, seed=None):
        """
        Args:
            motion_range (float, Tuple[float, float]): The range of motion difficulty.
            scheduler (CfgNode, optional): Config detailing scheduler.
            seed (int, optional): A seed for reproducibility.
        """
        if isinstance(motion_range, (float, int)):
            motion_range = (motion_range,)
        elif len(motion_range) > 2:
            raise ValueError("`motion_range` must have 2 or fewer values")

        self.warmup_method = None
        self.warmup_iters = 0
        if scheduler is not None:
            self.warmup_method = scheduler.WARMUP_METHOD
            self.warmup_iters = scheduler.WARMUP_ITERS
        self.motion_range = motion_range

        g = torch.Generator()
        if seed:
            g = g.manual_seed(seed)
        self.generator = g

    def choose_motion_range(self):
        """Chooses motion range based on warmup."""
        if not isinstance(self.motion_range, Sequence):
            return self.motion_range
        elif len(self.motion_range) == 1:
            return self.motion_range[0]

        if self.warmup_method:
            curr_iter = get_event_storage().iter
            warmup_iters = self.warmup_iters
            if self.warmup_method == "linear":
                motion_range = (
                    curr_iter / warmup_iters * (self.motion_range[1] - self.motion_range[0])
                )
            else:
                raise ValueError(f"`warmup_method={self.warmup_method}` not supported")
        else:
            motion_range = self.motion_range[1] - self.motion_range[0]

        g = self.generator
        motion_range = (
            self.motion_range[0] + motion_range * torch.rand(1, generator=g, device=g.device).item()
        )
        return motion_range

    def __call__(self, *args, **kwargs):
        return self.forward(*args, **kwargs)

    def forward(self, kspace, seed=None, clone=True) -> torch.Tensor:
        """Performs motion corruption on kspace image.

        Args:
            kspace (torch.Tensor): The complex tensor. Shape ``(N, Y, X, #coils, [2])``.
            seed (int, optional): Fixed seed at runtime (useful for generating testing vals).
            clone (bool, optional): If ``True``, return a cloned tensor.

        Returns:
            torch.Tensor: The motion corrupted kspace.

        Note:
            For backwards compatibility with torch<1.6, complex tensors may also have the shape
            ``(..., 2)``, where the 2 channels in the last dimension are real and
            imaginary, respectively.
        """
        # is_complex = False
        if clone:
            kspace = kspace.clone()
        phase_matrix = torch.zeros(kspace.shape, dtype=torch.complex64)
        width = kspace.shape[2]
        g = self.generator if seed is None else torch.Generator().manual_seed(seed)
        scale = self.choose_motion_range()

        odd_err = (2 * np.pi * scale) * torch.rand(1, generator=g).numpy() - np.pi * scale
        even_err = (2 * np.pi * scale) * torch.rand(1, generator=g).numpy() - np.pi * scale
        for line in range(width):
            if line % 2 == 0:
                rand_err = even_err
            else:
                rand_err = odd_err
            phase_error = torch.from_numpy(np.exp(-1j * rand_err))
            phase_matrix[:, :, line] = phase_error
        aug_kspace = kspace * phase_matrix
        return aug_kspace

    @classmethod
    def from_cfg(cls, cfg, seed=None, **kwargs):
        cfg = cfg.MODEL.CONSISTENCY.AUG.MOTION
        return cls(cfg.RANGE, scheduler=cfg.SCHEDULER, seed=seed, **kwargs)



from meddlr.data import build, catalog, collate, data_utils, slice_dataset
from meddlr.data.build import build_recon_train_loader, build_recon_val_loader  # noqa
from meddlr.data.catalog import DatasetCatalog, MetadataCatalog  # noqa
from meddlr.data.collate import collate_by_supervision, default_collate  # noqa: F401
from meddlr.data.datasets import *  # noqa
from meddlr.data.samplers import *  # noqa

__all__ = []
__all__.extend(build.__all__)
__all__.extend(collate.__all__)
__all__.extend(catalog.__all__)
__all__.extend(data_utils.__all__)
__all__.extend(slice_dataset.__all__)
